from util import *

REPS = 10

check_call("rm -f addstats delstats", shell=True)
for i in range(0, REPS):
    cont, ip = create_container("command time -o addstats -a -f '%e,%S,%U,%M,%t,%K,%I,%O' ")
    delete_container(cont, "command time -o delstats -a -f '%e,%S,%U,%M,%t,%K,%I,%O' ")

print "Add stats"
print "wallclock(s),system(s),user(s),max rss(KB),avg rss(KB),avg tot mem(KB),I,O"
check_call("cat addstats", shell=True)

print "\nDel stats"
print "wallclock(s),system(s),user(s),max rss(KB),avg rss(KB),avg tot mem(KB),I,O"
check_call("cat delstats", shell=True)



# Copyright (c) 2015-2016 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from unittest import TestCase

from tests.st.utils.utils import (get_ip, wipe_etcd, calicoctl)

HOST_IPV4 = get_ip()

logging.basicConfig(level=logging.DEBUG, format="%(message)s")
logger = logging.getLogger(__name__)


class TestBase(TestCase):
    """
    Base class for test-wide methods.
    """

    def setUp(self):
        """
        Clean up before every test.
        """
        self.ip = HOST_IPV4
        self.wipe_etcd()

        # Log a newline to ensure that the first log appears on its own line.
        logger.info("")

    def wipe_etcd(self):
        wipe_etcd(self.ip)



# -*- coding: utf-8 -*-
# Copyright (c) 2018 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Shared etcd data-model definitions for version 2 of the data model.

Version 2 is similar to version 1 but with region-awareness added into
some of the etcd key paths.

The wider Calico data model is now v3.  The legacy definitions here
are private to networking-calico (i.e. the Neutron driver and the
Calico DHCP agent) and Felix's status-reporting code.  However, when
changing these, we still need to consider upgrading an existing
Calico/OpenStack deployment.

"""
import re

from networking_calico import datamodel_v1

# Region string used when no region name has been configured.
NO_REGION = "no-region"

# Region string prefix used when a region name has been configured.
REGION_PREFIX = "region-"


# Subtree used by Felix to report its own status (as an OpenStack
# 'agent') and the status of each endpoint (or OpenStack 'port') that
# it is responsible for.
#
# Agent status is at /calico/felix/v2/<region_string>/host/<hostname>/status.
#
# Port status is at /calico/felix/v2/<region_string>//host/<hostname>/
#                    workload/openstack/<workload>/endpoint/<endpoint>.
def felix_status_dir(region_string=NO_REGION):
    return "/calico/felix/v2/%s/host" % region_string


# Regex to match endpoints, captures "hostname" and "endpoint_id".
# Works for endpoint status paths.
_cached_endpoint_key_re = None


def get_endpoint_id_from_key(region_string, key):
    global _cached_endpoint_key_re
    if _cached_endpoint_key_re is None:
        _cached_endpoint_key_re = re.compile(
            r"^(?:" + felix_status_dir(region_string) + r")"
            r"/(?P<hostname>[^/]+)/"
            r"workload/"
            r"(?P<orchestrator>[^/]+)/"
            r"(?P<workload_id>[^/]+)/"
            r"endpoint/(?P<endpoint_id>[^/]+)"
        )
    m = _cached_endpoint_key_re.match(key)
    if m:
        # Got an endpoint.
        host = m.group("hostname")
        orch = m.group("orchestrator")
        workload_id = m.group("workload_id")
        endpoint_id = m.group("endpoint_id")
        combined_id = datamodel_v1.WloadEndpointId(host, orch, workload_id, endpoint_id)
        return combined_id
    else:
        return None


def _reset_globals():
    global _cached_endpoint_key_re
    _cached_endpoint_key_re = None


# Region-aware subnet path.
def subnet_dir(region_string=NO_REGION):
    return "/calico/dhcp/v2/%s/subnet" % region_string


def key_for_subnet(subnet_id, region_string):
    return subnet_dir(region_string) + "/%s" % subnet_id


# Key used for leader election by Neutron mechanism drivers.
def neutron_election_key(region_string):
    return "/calico/openstack/v2/%s/neutron_election" % region_string



# Copyright (c) 2021 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import json
import subprocess
import time

from tests.k8st.test_base import TestBase
from tests.k8st.utils.utils import retry_until_success, kubectl, node_info

_log = logging.getLogger(__name__)

def create_status(name, node, interval):
    kubectl("""apply -f - <<EOF
apiVersion: projectcalico.org/v3
kind: CalicoNodeStatus
metadata:
    name: %s
spec:
    node: %s
    classes:
    - Agent
    - BGP
    - Routes
    updatePeriodSeconds: %d
EOF
""" % (name, node, interval))

def read_status(name):
    status_json = kubectl("get caliconodestatus %s -o json" % name)
    status_dict = json.loads(status_json)
    return status_dict['status']

def delete_status(name):
    kubectl("delete caliconodestatus %s" % name)

def is_subdict(small, big):
    return dict(big, **small) == big

class TestNodeMeshStatus(TestBase):
    def setUp(self):
        TestBase.setUp(self)

        # Get 2 worker node names. The first name
        # returned is always the master.
        self.nodes, self.ips, self.ip6s = node_info()

        self.test_node = self.nodes[1]
        self.test_node_ip = self.ips[1]
        self.test_node_ip6s = self.ip6s[1]
        self.status_name = "node-status-0"
        create_status(self.status_name, self.test_node, 10)

    def tearDown(self):
        # Delete resource
        delete_status(self.status_name)

    def test_dual_stack_status(self):
        retry_until_success(lambda: read_status(self.status_name), retries=5, wait_time=1)
        status = read_status(self.status_name)

        # Should have correct agent status.
        self.assert_is_subdict({'state': 'Ready', 'routerID': self.test_node_ip}, status['agent']['birdV4'])
        self.assert_is_subdict({'state': 'Ready', 'routerID': self.test_node_ip}, status['agent']['birdV6'])

        # Should have correct BGP status.
        self.assert_is_subdict({'numberEstablishedV4': 3, 'numberEstablishedV6': 3,
                                'numberNotEstablishedV4': 0, 'numberNotEstablishedV6': 0}, status['bgp'])

        # Should have correct peers and routes status.
        for i in [0, 2, 3]:
            peers = status['bgp']['peersV4']
            self.assert_dictlist_has_subdict({'peerIP': self.ips[i], 'state': 'Established', 'type': 'NodeMesh'}, peers)

            peers = status['bgp']['peersV6']
            self.assert_dictlist_has_subdict({'peerIP': self.ip6s[i], 'state': 'Established', 'type': 'NodeMesh'}, peers)

            # Should have correct route status.
            routes = status['routes']['routesV4']
            self.assert_dictlist_has_subdict({"gateway": self.ips[i], "interface": "eth0", 'type': 'FIB',
                                          "learnedFrom": {
                                              "sourceType": "NodeMesh"
                                          }}, routes)

            routes = status['routes']['routesV6']
            self.assert_dictlist_has_subdict({"gateway": self.ip6s[i], "interface": "eth0", 'type': 'FIB',
                                          "learnedFrom": {
                                              "sourceType": "NodeMesh"
                                          }}, routes)

    def assert_is_subdict(self, small, big):
        self.assertTrue(is_subdict(small, big))

    def assert_dictlist_has_subdict(self, small, bigList):
        for big in bigList:
            if is_subdict(small, big):
                return True
        return False

'''
Sample output of caliconodestatus
{
    "apiVersion": "projectcalico.org/v3",
    "kind": "CalicoNodeStatus",
    "metadata": {
        "annotations": {
            "kubectl.kubernetes.io/last-applied-configuration": "{\"apiVersion\":\"projectcalico.org/v3\",\"kind\":\"CalicoNodeStatus\",\"metadata\":{\"annotations\":{},\"name\":\"node-status-0\"},\"spec\":{\"classes\":[\"Agent\",\"BGP\",\"Routes\"],\"node\":\"kind-worker\",\"updatePeriodSeconds\":10}}\n"
        },
        "creationTimestamp": "2021-11-18T23:04:58Z",
        "name": "node-status-0",
        "resourceVersion": "396418",
        "uid": "e306e68e-d7d7-46fb-80a2-d111676012fa"
    },
    "spec": {
        "classes": [
            "Agent",
            "BGP",
            "Routes"
        ],
        "node": "kind-worker",
        "updatePeriodSeconds": 10
    },
    "status": {
        "agent": {
            "birdV4": {
                "lastBootTime": "2021-11-18 23:06:36",
                "lastReconfigurationTime": "2021-11-18 23:06:36",
                "routerID": "172.18.0.2",
                "state": "Ready",
                "version": "v0.3.3+birdv1.6.8"
            },
            "birdV6": {
                "lastBootTime": "2021-11-18 23:06:36",
                "lastReconfigurationTime": "2021-11-18 23:06:36",
                "routerID": "172.18.0.2",
                "state": "Ready",
                "version": "v0.3.3+birdv1.6.8"
            }
        },
        "bgp": {
            "numberEstablishedV4": 3,
            "numberEstablishedV6": 3,
            "numberNotEstablishedV4": 0,
            "numberNotEstablishedV6": 0,
            "peersV4": [
                {
                    "peerIP": "172.18.0.3",
                    "since": "23:06:37",
                    "state": "Established",
                    "type": "NodeMesh"
                },
                {
                    "peerIP": "172.18.0.4",
                    "since": "23:06:37",
                    "state": "Established",
                    "type": "NodeMesh"
                },
                {
                    "peerIP": "172.18.0.5",
                    "since": "23:06:36",
                    "state": "Established",
                    "type": "NodeMesh"
                }
            ],
            "peersV6": [
                {
                    "peerIP": "2001:20::8",
                    "since": "23:06:36",
                    "state": "Established",
                    "type": "NodeMesh"
                },
                {
                    "peerIP": "2001:20::2",
                    "since": "23:06:37",
                    "state": "Established",
                    "type": "NodeMesh"
                },
                {
                    "peerIP": "2001:20::3",
                    "since": "23:06:36",
                    "state": "Established",
                    "type": "NodeMesh"
                }
            ]
        },
        "lastUpdated": "2021-11-18T23:07:05Z",
        "routes": {
            "routesV4": [
                {
                    "destination": "0.0.0.0/0",
                    "gateway": "172.18.0.1",
                    "interface": "eth0",
                    "learnedFrom": {
                        "sourceType": "Kernel"
                    },
                    "type": "FIB"
                },
                {
                    "destination": "192.168.110.128/26",
                    "gateway": "172.18.0.4",
                    "interface": "eth0",
                    "learnedFrom": {
                        "peerIP": "172.18.0.4",
                        "sourceType": "NodeMesh"
                    },
                    "type": "FIB"
                },
                {
                    "destination": "192.168.82.0/26",
                    "gateway": "172.18.0.3",
                    "interface": "eth0",
                    "learnedFrom": {
                        "peerIP": "172.18.0.3",
                        "sourceType": "NodeMesh"
                    },
                    "type": "FIB"
                },
                {
                    "destination": "192.168.82.0/26",
                    "gateway": "172.18.0.3",
                    "interface": "eth0",
                    "learnedFrom": {
                        "peerIP": "172.18.0.3",
                        "sourceType": "NodeMesh"
                    },
                    "type": "RIB"
                },
                {
                    "destination": "192.168.195.192/26",
                    "gateway": "172.18.0.5",
                    "interface": "eth0",
                    "learnedFrom": {
                        "peerIP": "172.18.0.5",
                        "sourceType": "NodeMesh"
                    },
                    "type": "FIB"
                },
                {
                    "destination": "192.168.195.192/26",
                    "gateway": "172.18.0.5",
                    "interface": "eth0",
                    "learnedFrom": {
                        "peerIP": "172.18.0.5",
                        "sourceType": "NodeMesh"
                    },
                    "type": "RIB"
                },
                {
                    "destination": "192.168.162.128/26",
                    "gateway": "N/A",
                    "interface": "blackhole",
                    "learnedFrom": {
                        "sourceType": "Static"
                    },
                    "type": "FIB"
                },
                {
                    "destination": "192.168.162.128/26",
                    "gateway": "N/A",
                    "interface": "blackhole",
                    "learnedFrom": {
                        "sourceType": "Kernel"
                    },
                    "type": "RIB"
                },
                {
                    "destination": "192.168.162.128/32",
                    "gateway": "N/A",
                    "interface": "calibb219784141",
                    "learnedFrom": {
                        "sourceType": "Kernel"
                    },
                    "type": "FIB"
                },
                {
                    "destination": "192.168.162.129/32",
                    "gateway": "N/A",
                    "interface": "calid15e5529ae8",
                    "learnedFrom": {
                        "sourceType": "Kernel"
                    },
                    "type": "FIB"
                },
                {
                    "destination": "192.168.162.130/32",
                    "gateway": "N/A",
                    "interface": "calia7a3e029a7b",
                    "learnedFrom": {
                        "sourceType": "Kernel"
                    },
                    "type": "FIB"
                },
                {
                    "destination": "192.168.162.132/32",
                    "gateway": "N/A",
                    "interface": "wireguard.cali",
                    "learnedFrom": {
                        "sourceType": "Direct"
                    },
                    "type": "FIB"
                },
                {
                    "destination": "172.18.0.0/16",
                    "gateway": "N/A",
                    "interface": "eth0",
                    "learnedFrom": {
                        "sourceType": "Direct"
                    },
                    "type": "FIB"
                }
            ],
            "routesV6": [
                {
                    "destination": "::/0",
                    "gateway": "fc00:f853:ccd:e793::1",
                    "interface": "eth0",
                    "learnedFrom": {
                        "sourceType": "Kernel"
                    },
                    "type": "FIB"
                },
                {
                    "destination": "fd00:10:244:0:586d:4461:e980:a282/128",
                    "gateway": "N/A",
                    "interface": "calia7a3e029a7b",
                    "learnedFrom": {
                        "sourceType": "Kernel"
                    },
                    "type": "FIB"
                },
                {
                    "destination": "fd00:10:244:0:586d:4461:e980:a280/122",
                    "gateway": "N/A",
                    "interface": "blackhole",
                    "learnedFrom": {
                        "sourceType": "Static"
                    },
                    "type": "FIB"
                },
                {
                    "destination": "fd00:10:244:0:586d:4461:e980:a280/128",
                    "gateway": "N/A",
                    "interface": "calibb219784141",
                    "learnedFrom": {
                        "sourceType": "Kernel"
                    },
                    "type": "FIB"
                },
                {
                    "destination": "fd00:10:244:0:586d:4461:e980:a281/128",
                    "gateway": "N/A",
                    "interface": "calid15e5529ae8",
                    "learnedFrom": {
                        "sourceType": "Kernel"
                    },
                    "type": "FIB"
                },
                {
                    "destination": "2001:20::/64",
                    "gateway": "N/A",
                    "interface": "eth0",
                    "learnedFrom": {
                        "sourceType": "Direct"
                    },
                    "type": "FIB"
                },
                {
                    "destination": "fc00:f853:ccd:e793::/64",
                    "gateway": "N/A",
                    "interface": "eth0",
                    "learnedFrom": {
                        "sourceType": "Direct"
                    },
                    "type": "FIB"
                },
                {
                    "destination": "fd00:10:244:0:58fd:b191:5c13:9cc0/122",
                    "gateway": "2001:20::3",
                    "interface": "eth0",
                    "learnedFrom": {
                        "peerIP": "2001:20::3",
                        "sourceType": "NodeMesh"
                    },
                    "type": "FIB"
                },
                {
                    "destination": "fd00:10:244:0:1cc0:b1ac:ad47:e7c0/122",
                    "gateway": "2001:20::2",
                    "interface": "eth0",
                    "learnedFrom": {
                        "peerIP": "2001:20::2",
                        "sourceType": "NodeMesh"
                    },
                    "type": "FIB"
                },
                {
                    "destination": "fd00:10:244:0:dec7:e00b:6677:b640/122",
                    "gateway": "2001:20::8",
                    "interface": "eth0",
                    "learnedFrom": {
                        "peerIP": "2001:20::8",
                        "sourceType": "NodeMesh"
                    },
                    "type": "FIB"
                }
            ]
        }
    }
}
'''



# -*- coding: utf-8 -*-
# Copyright (c) 2018-2025 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from importlib.metadata import version

from etcd3gw.client import Etcd3Client
from etcd3gw.exceptions import Etcd3Exception
from etcd3gw.lease import Lease
from etcd3gw.utils import _encode
from etcd3gw.utils import _increment_last_byte

from oslo_config import cfg

from oslo_log import log

from packaging.version import Version

# Incantations for enabling oslo_log debug logging, when desired:
# log.register_options(cfg.CONF)
# cfg.CONF.debug = True
# cfg.CONF.use_stderr = True
# log.setup(cfg.CONF, "demo")

LOG = log.getLogger(__name__)

# Limit on number of keys we get from etcd.  We found that the etcd gateway
# has limits on the size of responses that kick in at 3000+ keys so make sure
# we leave plenty of headroom.
CHUNK_SIZE_LIMIT = 200

# Indicates that a put operation must update an existing resource and not
# create a new resource.
MUST_UPDATE = "MUST_UPDATE"


class KeyNotFound(Etcd3Exception):
    pass


def get(key, with_lease=False):
    """Read a value from etcdv3.

    - key (string): The key to read.

    - with_lease (boolean): Indicates also to return the key's lease.

    Returns (value, mod_revision) or (value, mod_revision, lease) where

    - value is the key's value

    - mod_revision is the etcdv3 revision at which the key was last
      modified

    - lease is an etcd3gw.lease.Lease object representing the key's lease, if
      it has one; or else None.

    Raises KeyNotFound if there is no resource with that kind and name.
    """
    client = _get_client()
    results = client.get(key, metadata=True)
    LOG.debug("etcdv3 get key=%s results=%s", key, results)
    if len(results) != 1:
        raise KeyNotFound()
    value, item = results[0]
    value = value.decode()
    if with_lease:
        lease = None
        if "lease" in item:
            lease = Lease(int(item["lease"]), client)
        return value, item["mod_revision"], lease
    else:
        return value, item["mod_revision"]


def put(key, value, mod_revision=None, lease=None, existing_value=None):
    """Write a key/value pair to etcdv3.

    - key (string): The key to write.

    - value (string): The value to write.

    - mod_revision (string): If specified, indicates that the write should only
      proceed if replacing an existing value with that mod_revision.
      mod_revision=0 indicates that the key must not yet exist, i.e. that this
      write will create it.

    - lease: If specified, a Lease object to associate with the key.

    - existing_value (string): If specified, indicates that the write should
      only proceed if replacing that existing value.

    Returns True if the write happened successfully; False if not.
    """
    client = _get_client()
    LOG.debug("etcdv3 put key=%s value=%s mod_revision=%r", key, value, mod_revision)
    txn = {}
    if mod_revision == 0:
        # Write operation must _create_ the KV entry.
        base64_key = _encode(key)
        txn["compare"] = [
            {
                "key": base64_key,
                "result": "EQUAL",
                "target": "VERSION",
                "version": 0,
            }
        ]
    elif mod_revision == MUST_UPDATE:
        # Write operation must update and _not_ create the KV entry.
        base64_key = _encode(key)
        txn["compare"] = [
            {
                "key": base64_key,
                "result": "NOT_EQUAL",
                "target": "VERSION",
                "version": 0,
            }
        ]
    elif mod_revision is not None:
        # Write operation must _replace_ a KV entry with the specified
        # revision.
        base64_key = _encode(key)
        txn["compare"] = [
            {
                "key": base64_key,
                "result": "EQUAL",
                "target": "MOD",
                "mod_revision": mod_revision,
            }
        ]
    elif existing_value is not None:
        # Write operation must _replace_ a KV entry with the specified value.
        base64_key = _encode(key)
        base64_existing = _encode(existing_value)
        txn["compare"] = [
            {
                "key": base64_key,
                "result": "EQUAL",
                "target": "VALUE",
                "value": base64_existing,
            }
        ]
    if txn:
        base64_value = _encode(value)
        txn["success"] = [
            {
                "request_put": {
                    "key": base64_key,
                    "value": base64_value,
                },
            }
        ]
        txn["failure"] = []
        if lease is not None:
            txn["success"][0]["request_put"]["lease"] = lease.id
        result = client.transaction(txn)
        LOG.debug("transaction result %s", result)
        succeeded = result.get("succeeded", False)
    else:
        succeeded = client.put(key, value, lease=lease)
    return succeeded


def delete(key, existing_value=None, mod_revision=None):
    """Delete a key/value pair from etcdv3.

    - key (string): The key to delete.

    - existing_value (string): If specified, indicates that the delete should
      only proceed if the existing value is this.

    - mod_revision (string): If specified, indicates that the delete should
      only proceed if deleting an existing value with that mod_revision.

    Returns True if the deletion was successful; False if not.
    """
    client = _get_client()
    LOG.debug("etcdv3 delete key=%s", key)
    if mod_revision is not None:
        base64_key = _encode(key)
        txn = {
            "compare": [
                {
                    "key": base64_key,
                    "result": "EQUAL",
                    "target": "MOD",
                    "mod_revision": mod_revision,
                }
            ],
            "success": [
                {
                    "request_delete_range": {
                        "key": base64_key,
                    },
                }
            ],
            "failure": [],
        }
        result = client.transaction(txn)
        LOG.debug("transaction result %s", result)
        deleted = result.get("succeeded", False)
    elif existing_value is not None:
        base64_key = _encode(key)
        base64_existing = _encode(existing_value)
        txn = {
            "compare": [
                {
                    "key": base64_key,
                    "result": "EQUAL",
                    "target": "VALUE",
                    "value": base64_existing,
                }
            ],
            "success": [
                {
                    "request_delete_range": {
                        "key": base64_key,
                    },
                }
            ],
            "failure": [],
        }
        result = client.transaction(txn)
        LOG.debug("transaction result %s", result)
        deleted = result.get("succeeded", False)
    else:
        deleted = client.delete(key)
    LOG.debug("etcdv3 deleted=%s", deleted)
    return deleted


def delete_prefix(prefix):
    """Best effort deletion of all keys beginning with PREFIX."""
    LOG.debug("etcdv3 delete_prefix prefix=%s", prefix)
    client = _get_client()
    return client.delete_prefix(prefix)


def get_prefix(prefix, revision=None):
    """Read all etcdv3 data whose key begins with a given prefix.

    - prefix (string): The prefix.

    - revision: The revision to do the get at.  If not specified then the
      current revision is used.

    Returns a list of tuples (key, value, mod_revision), one for each key-value
    pair, in which:

    - key is the etcd key (a string)

    - value is the etcd value (also a string; note *not* JSON-decoded)

    - mod_revision is the revision at which that key was last modified (an
      integer represented as a string).

    Note: this entrypoint is only used for data outside the Calico v3 data
    model; specifically for legacy Calico v1 status notifications.  This
    entrypoint should be removed once those status notifications have been
    reimplemented within the Calico v3 data model.
    """
    client = _get_client()

    if revision is None:
        _, revision = get_status()
        LOG.debug("Doing get at current revision: %r", revision)

    # The JSON gateway can only return a certain number of bytes in a single
    # response so we chunk up the read into blocks.
    #
    # Since etcd's get protocol has an inclusive range_start and an exclusive
    # range_end, we load the keys in reverse order.  That way, we can use the
    # final key in each chunk as the next range_end.
    range_end = _encode(_increment_last_byte(prefix))
    results = []
    while True:
        # Note: originally, we included the sort_target parameter here but
        # etcdgw has a bug (https://github.com/dims/etcd3-gateway/issues/18),
        # which prevents that from working.  In any case, sort-by-key is the
        # default, which is what we want.
        chunk = client.get(
            prefix,
            metadata=True,
            range_end=range_end,
            sort_order="descend",
            limit=CHUNK_SIZE_LIMIT,
            revision=str(revision),
        )
        results.extend(chunk)
        if len(chunk) < CHUNK_SIZE_LIMIT:
            # Partial (or empty) chunk signals that we're done.
            break
        _, data = chunk[-1]
        range_end = _encode(data["key"])

    LOG.debug("etcdv3 get_prefix %s results=%s", prefix, len(results))
    tuples = []
    for result in results:
        value, item = result
        t = (item["key"].decode(), value.decode(), item["mod_revision"])
        tuples.append(t)
    return tuples


def watch_subtree(prefix, start_revision):
    """Watch for changes to etcdv3 data whose key begins with a given prefix.

    - prefix (string): The prefix.

    - start_revision (string representation of an integer): The revision to
      start watching from.  Events will be reported beginning from, and
      including, this revision.

    Returns a tuple (event_stream, cancel), in which:

    - event_stream is a generator that returns the next reported event, or None
      if cancel has been called.  Each event is a dict like

      {'kv': {'key': <string>,
              'value': <string>,
              'mod_revision': <string>,
              'create_revision': <string>,
              'version': <string>}}

      or

      {'type': 'DELETE',
       'kv': {'key': <key>,
              'mod_revision': <string>}}

    - cancel is a thunk that can be called to cancel the watch and cause the
      event_stream to return None.

    Note: this entrypoint is only used for data outside the Calico v3 data
    model; specifically for legacy Calico v1 status notifications.  This
    entrypoint should be updated or removed when those status notifications
    have been reimplemented within the Calico v3 data model.
    """
    LOG.debug("Watch subtree %s from revision %r", prefix, start_revision)
    client = _get_client()
    event_stream, cancel = client.watch_prefix(prefix, start_revision=start_revision)
    return event_stream, cancel


def get_status():
    """Get the current etcdv3 cluster ID and revision.

    Returns a tuple (cluster_id, revision).
    """
    client = _get_client()
    status = client.status()
    LOG.debug("etcdv3 status %s", status)
    return status["header"]["cluster_id"], status["header"]["revision"]


def request_compaction(revision):
    """Request compaction at the specified revision."""
    client = _get_client()
    LOG.debug("request etcdv3 compaction at %r", revision)
    response = client.post(
        client.get_url("/kv/compaction"), json={"revision": str(revision)}
    )
    LOG.debug("=> %s", response)


def watch_once(key, timeout=None, **kwargs):
    """Watch a key and stop after the first event.

    :param key: key to watch
    :param timeout: (optional) timeout in seconds.
    :returns: event
    """
    client = _get_client()
    LOG.debug("etcdv3 watch_once %s timeout %r kwargs %r", key, timeout, kwargs)
    return client.watch_once(key, timeout=timeout, **kwargs)


def get_lease(ttl):
    """Get a lease for the specified TTL (in seconds)."""
    client = _get_client()
    return client.lease(ttl=ttl)


def logging_exceptions(fn):
    """Decorator to log (and reraise) Etcd3Exceptions."""

    @functools.wraps(fn)
    def wrapped(self, *args, **kwargs):
        try:
            return fn(self, *args, **kwargs)
        except Etcd3Exception as e:
            LOG.warning("Etcd3Exception, re-raising: %r:\n%s", e, e.detail_text)
            raise

    return wrapped


# Internals.
_client = None


# Possible API paths for connecting to an etcd server.  Defined as a variable
# here so that test code can override it after importing this file.
_possible_etcd_api_paths = ["/v3/", "/v3beta/", "/v3alpha/"]


# Wrap Etcd3Client to authenticate when needed and add an
# Authorization header to the session headers.
#
# All of networking-calico's etcd operations go through either (1)
# etcd3gw.client.Etcd3Client.post, or (2) etcd3gw.watch.Watcher.
# Here, we hook (1) so as to authenticate when the normal POST request
# fails; then we add the returned auth token as an Authorization
# header on the underlying session.  Adding that header to the session
# means that it will apply to etcd3gw.watch.Watcher, as well as to
# POST requests for all non-watch operations.
#
# The question arises what happens if we do a watch operation before
# there is correct Authorization on the session?  Firstly this is
# unlikely, because we always watch following a preceding get of the
# same key or subtree; but it might still be possible if the etcd
# server is restarted after the get and before the watch.  Secondly,
# we handle the watch request failing and loop round to retry the get
# again.  Overall, therefore, we are safe on this point.
class Etcd3AuthClient(Etcd3Client):
    def __init__(
        self,
        host="localhost",
        port=2379,
        protocol="http",
        ca_cert=None,
        cert_key=None,
        cert_cert=None,
        timeout=None,
        username=None,
        password=None,
    ):
        global _possible_etcd_api_paths
        possible_api_paths = _possible_etcd_api_paths
        created_working_client = False
        while not created_working_client:
            try:
                LOG.info("Try creating etcd3gw client with %s", possible_api_paths[0])
                super(Etcd3AuthClient, self).__init__(
                    host=host,
                    port=port,
                    protocol=protocol,
                    ca_cert=ca_cert,
                    cert_key=cert_key,
                    cert_cert=cert_cert,
                    timeout=timeout,
                    api_path=possible_api_paths[0],
                )
                possible_api_paths = possible_api_paths[1:]
            except TypeError:
                # Indicates an old version of etcd3gw that doesn't support the
                # api_path keyword.
                possible_api_paths = []
                super(Etcd3AuthClient, self).__init__(
                    host=host,
                    port=port,
                    protocol=protocol,
                    ca_cert=ca_cert,
                    cert_key=cert_key,
                    cert_cert=cert_cert,
                    timeout=timeout,
                )

            self.username = username
            self.password = password

            # Now test if this client is really working.
            try:
                status = self.status()
                LOG.info("Status = %r", status)
                created_working_client = True
            except Exception:
                LOG.exception("etcd3gw client not working")
                # If there are no more possible API paths to try, reraise the
                # current exception.
                if not possible_api_paths:
                    raise

    def authenticate(self):
        # When authenticating, there mustn't be an Authorization
        # header with an old token, or else etcd responds with
        # "Unauthorized: invalid auth token".  So remove any existing
        # Authorization header.
        if "Authorization" in self.session.headers:
            del self.session.headers["Authorization"]

        # Send authenticate request.  If this raises an exception,
        # e.g. because of a connectivity issue to the etcd server,
        # it's OK for that to bubble up and be handled in the code
        # that called post.
        response = super(Etcd3AuthClient, self).post(
            self.get_url("/auth/authenticate"),
            json={"name": self.username, "password": self.password},
        )

        # Add Authorization header with the received token to the
        # underlying requests session.  This covers all subsequent
        # requests, and is needed in particular for watches, because
        # the watch code does not use client.post and so could not be
        # covered by adding a header to kwargs in the following post
        # method.
        self.session.headers["Authorization"] = response["token"]

    def post(self, *args, **kwargs):
        if Version(version("etcd3gw")) < Version("2.4.0"):
            # Impose a maximum timeout, according to the [calico]
            # etcd_timeout config parameter.  Imposing a timeout is
            # generally a good idea, and specifically we want to protect
            # this code from the apparent etcdserver hang bug at
            # https://github.com/etcd-io/etcd/issues/11377.
            if (
                "timeout" not in kwargs
                or kwargs["timeout"] > cfg.CONF.calico.etcd_timeout
            ):
                kwargs["timeout"] = cfg.CONF.calico.etcd_timeout
        try:
            # Try the post.  If no authentication is needed, or if an
            # Authorization token has been added to the session's
            # headers, and is still valid, this should succeed.
            return super(Etcd3AuthClient, self).post(*args, **kwargs)
        except Etcd3Exception as e:
            if self.username and self.password:
                # Etcd auth credentials are configured, so assume the
                # problem might be that we need to authenticate or
                # re-authenticate.
                LOG.info("Might need to (re)authenticate: %r:\n%s", e, e.detail_text)

                # Authenticate and then reissue the request.
                self.authenticate()
                return super(Etcd3AuthClient, self).post(*args, **kwargs)

            # Otherwise re-raise.
            raise


def _get_client():
    global _client
    if not _client:
        calico_cfg = cfg.CONF.calico
        tls_config_params = [
            calico_cfg.etcd_key_file,
            calico_cfg.etcd_cert_file,
            calico_cfg.etcd_ca_cert_file,
        ]
        # Impose a maximum timeout, according to the [calico] etcd_timeout
        # config parameter.  Imposing a timeout is generally a good idea, and
        # specifically we want to protect this code from the apparent
        # etcdserver hang bug at https://github.com/etcd-io/etcd/issues/11377.
        # etcd3gw 1.0.1 and 2.4.0 both allow specifying a timeout on the
        # following client constructors, but the timeout only actually works in
        # 2.4.0 onwards.  For etcd3gw<=2.4.0 we work around that by also
        # specifying the timeout as a kwarg on each post() call.
        if any(tls_config_params):
            LOG.info(
                "TLS to etcd is enabled with key file %s; "
                "cert file %s; CA cert file %s",
                *tls_config_params
            )
            _client = Etcd3AuthClient(
                host=calico_cfg.etcd_host,
                port=calico_cfg.etcd_port,
                timeout=cfg.CONF.calico.etcd_timeout,
                protocol="https",
                ca_cert=calico_cfg.etcd_ca_cert_file,
                cert_key=calico_cfg.etcd_key_file,
                cert_cert=calico_cfg.etcd_cert_file,
                username=calico_cfg.etcd_username,
                password=calico_cfg.etcd_password,
            )
        else:
            LOG.info("TLS disabled, using HTTP to connect to etcd.")
            _client = Etcd3AuthClient(
                host=calico_cfg.etcd_host,
                port=calico_cfg.etcd_port,
                timeout=cfg.CONF.calico.etcd_timeout,
                protocol="http",
                username=calico_cfg.etcd_username,
                password=calico_cfg.etcd_password,
            )
    return _client



from util import *

cont1, ip = create_container()
cont2, _ = create_container()

run_command(cont1, "date")
run_command(cont1, "sleep 2")

# The ping will only work if felix is running
run_command(cont2, "ping %s" % ip)

delete_container(cont1)
delete_container(cont2)




# Copyright (c) 2015-2019 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from tests.st.test_base import TestBase
from tests.st.utils.data import node_name1_rev1
from tests.st.utils.utils import calicoctl, name, set_cluster_version

class TestCalicoctlCLIFlags(TestBase):
    """
    Test calicoctl command-line flags
    """

    def test_usage(self):
        """
        Test usage successfully
        """
        rc = calicoctl("-h")
        rc.assert_no_error()

    def test_bad_flags(self):
        """
        Test bad flags unsuccessfully
        """
        # no -m flag
        rc = calicoctl("-m")
        rc.assert_error(text="Invalid option: '--allow-version-mismatch -m'.")
        # no --unknown-flag flag
        rc = calicoctl("--unknown-flag")
        rc.assert_error(text="Invalid option: '--allow-version-mismatch --unknown-flag'.")

    def test_version_mismatch(self):
        """
        Test version mismatch verification
        """
        # Create a Node, this should also trigger auto-creation of a cluster info,
        # but without any CalicoVersion field.
        rc = calicoctl("create", data=node_name1_rev1)
        rc.assert_no_error()

        # The "datastore migrate import" command bypasses version mismatch checking
        rc = calicoctl("datastore migrate import -f a", allowVersionMismatch=False)
        # Assert that the error is not "version mismatch"
        rc.assert_error("invalid datastore type")

        # CalicoVersion is unset in the cluster, expect no error
        rc = calicoctl("replace", data=node_name1_rev1, allowVersionMismatch=False)
        rc.assert_no_error()

        # Set the correct CalicoVersion in the cluster
        output = set_cluster_version()
        output.assert_no_error()

        # CalicoVersion is correct in the cluster, expect no error
        rc = calicoctl("replace", data=node_name1_rev1, allowVersionMismatch=False)
        rc.assert_no_error()

        # Set an incorrect CalicoVersion in the cluster
        rc = set_cluster_version("v0.0.0.1.2.3")
        rc.assert_no_error()

        # CalicoVersion is incorrect in the cluster, expect error
        rc = calicoctl("replace", data=node_name1_rev1, allowVersionMismatch=False)
        rc.assert_error("version mismatch.")



# -*- coding: utf-8 -*-
# Copyright (c) 2018 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import re
import uuid

from oslo_log import log

from networking_calico import datamodel_v2
from networking_calico import etcdv3
from networking_calico.timestamp import timestamp_now

# Particular JSON key strings.
CLUSTER_GUID = "clusterGUID"
CLUSTER_TYPE = "clusterType"
DATASTORE_READY = "datastoreReady"
ENDPOINT_REPORTING_ENABLED = "endpointReportingEnabled"
INTERFACE_PREFIX = "interfacePrefix"

# Annotation keys.
ANN_KEY_PREFIX = "openstack.projectcalico.org/"
ANN_KEY_FQDN = ANN_KEY_PREFIX + "fqdn"
ANN_KEY_NETWORK_ID = ANN_KEY_PREFIX + "network-id"

# Namespace constants.
NO_REGION_NAMESPACE = "openstack"
REGION_NAMESPACE_PREFIX = "openstack-"
NOT_NAMESPACED = None


LOG = log.getLogger(__name__)


def put(
    resource_kind, namespace, name, spec, annotations={}, labels=None, mod_revision=None
):
    """Write a Calico v3 resource to etcdv3.

    - resource_kind (string): E.g. WorkloadEndpoint, Profile, etc.

    - name (string): The resource's name.  This is used to form its etcd key,
      and also goes in its .Metadata.Name field.

    - namespace (string): The namespace to put the resource in.

    - spec (dict): Resource spec, as a dict with keys as specified by the
      'json:' comments in the relevant golang struct definition (for example,
      https://github.com/projectcalico/libcalico-go/blob/master/
      lib/apis/v3/workloadendpoint.go#L38).

    - annotations (dict): Annotations to set on the resource.  These are merged
      with existing annotations; i.e. existing annotations with other keys are
      unchanged, and existing annotations with the same keys are overwritten by
      these new values.

    - mod_revision: If 0 (a number), indicates that the write should only
      proceed if it _creates_ the resource.  Else, if etcdv3.MUST_UPDATE,
      indicates that the write should only proceed if it _does not create_ the
      resource.  Else, if not None, must be a number encoded as a string,
      e.g. "12345", and indicates that the write should only proceed if
      replacing an existing value with that mod_revision.

    Returns the resource's metadata UID (a string) if the write happened
    successfully; None if not.  The returned UID is truthy on success, so
    callers that previously checked truthiness are unaffected.
    """
    key = _build_key(resource_kind, namespace, name)
    value = None
    try:
        # Get the existing resource so we can persist its metadata.
        value, _ = _get_with_metadata(resource_kind, namespace, name)
    except etcdv3.KeyNotFound:
        pass
    except ValueError:
        LOG.warning("etcd value not valid JSON, so ignoring")
    if value is None:
        # Build basic resource structure.
        value = {
            "kind": resource_kind,
            "apiVersion": "projectcalico.org/v3",
            "metadata": {
                "name": name,
            },
        }
    # Ensure namespace set, for a namespaced resource.
    if _is_namespaced(resource_kind):
        assert namespace is not None
        value["metadata"]["namespace"] = namespace
    # Ensure that there is a creation timestamp.
    if "creationTimestamp" not in value["metadata"]:
        value["metadata"]["creationTimestamp"] = timestamp_now()
    # Ensure that there is a UID.
    if "uid" not in value["metadata"]:
        value["metadata"]["uid"] = str(uuid.uuid4())
    # Set annotations and labels if specified.  (We previously used to merge
    # here, instead of overwriting, but (a) for annotations there is actually
    # no use case for that, because we only use annotations on endpoints for
    # which Neutron is the sole source of truth; and (b) for the use case where
    # labels are used to represent security group membership it is crucial that
    # we overwrite and don't merge; otherwise a VM could never be removed from
    # a security group.)
    if annotations:
        value["metadata"]["annotations"] = annotations
    if labels:
        value["metadata"]["labels"] = labels
    # Set the new spec (overriding whatever may already be there).
    value["spec"] = spec
    uid = value["metadata"]["uid"]
    if etcdv3.put(key, json.dumps(value), mod_revision=mod_revision):
        return uid
    return None


def get(resource_kind, name):
    """Read spec of a non-namespaced Calico v3 resource from etcdv3.

    - resource_kind (string): E.g. ClusterInformation.

    - name (string): The resource's name, which is used to form its etcd key.

    Returns (spec, mod_revision) where

    - spec is the resource spec as a dict with keys as specified by the 'json:'
      comments in the relevant golang struct definition (for example,
      https://github.com/projectcalico/libcalico-go/blob/master/
      lib/apis/v3/clusterinfo.go#L39).

    - mod_revision is the etcdv3 revision at which the resource was last
      modified.

    Raises etcdv3.KeyNotFound if there is no resource with that kind and name.
    """
    value, mod_revision = _get_with_metadata(resource_kind, NOT_NAMESPACED, name)
    return value["spec"], mod_revision


def delete_legacy(resource_kind, name_prefix=""):
    key = _build_key(resource_kind, NO_REGION_NAMESPACE, name_prefix)
    etcdv3.delete_prefix(key)


def get_all(resource_kind, namespace, with_labels_and_annotations=False, revision=None):
    """Read all Calico v3 resources of a certain kind from etcdv3.

    - resource_kind (string): E.g. WorkloadEndpoint, Profile, etc.

    - namespace (string): The namespace to get resources for.

    - with_labels_and_annotations: If True, indicates to return the labels and
      annotations for each resource as well as the spec.

    - revision: if specified, the get is performed at the given revision
      as a snapshot.

    Returns a list of tuples (name, spec, mod_revision) or (name, (spec,
    labels, annotations), mod_revision), one for each resource of the specified
    kind, in which:

    - name is the resource's name (a string)

    - spec is a dict with keys as specified by the 'json:' comments in the
      relevant golang struct definition (for example,
      https://github.com/projectcalico/libcalico-go/blob/master/
      lib/apis/v3/workloadendpoint.go#L38).

    - labels is a dict containing the resource's labels

    - annotations is a dict containing the resource's annotations

    - mod_revision is the revision at which that resource was last modified (an
      integer represented as a string).
    """
    prefix = _build_key(resource_kind, namespace, "")
    results = etcdv3.get_prefix(prefix, revision=revision)
    tuples = []
    for result in results:
        key, value, mod_revision = result
        name = key.split("/")[-1]

        # Decode the value.
        spec = labels = annotations = None
        try:
            value_dict = json.loads(value)
            LOG.debug("value dict: %s", value_dict)
            spec = value_dict["spec"]
            labels = value_dict["metadata"].get("labels", {})
            annotations = value_dict["metadata"].get("annotations", {})
        except ValueError:
            # When the value is not valid JSON, we still return a tuple for
            # this key, with spec, labels and annotations all as None.  This is
            # so that the caller can correctly differentiate between
            # overwriting an existing value (which => a transaction with
            # specified mod_revision) and creating a key that did not exist
            # before (=> a transaction with version 0).
            LOG.warning("etcd value not valid JSON (%s)", value)

        if with_labels_and_annotations:
            t = (name, (spec, labels, annotations), mod_revision)
        else:
            t = (name, spec, mod_revision)
        tuples.append(t)
    return tuples


def delete(resource_kind, namespace, name, mod_revision=None):
    """Delete a Calico v3 resource from etcdv3.

    - resource_kind (string): E.g. WorkloadEndpoint, Profile, etc.

    - namespace (string): The namespace to delete the resource in.

    - name (string): The resource's name, which is used to form its etcd key.

    Returns True if the deletion was successful; False if not.
    """
    key = _build_key(resource_kind, namespace, name)
    return etcdv3.delete(key, mod_revision=mod_revision)


def get_uid(resource_kind, namespace, name):
    """Read the metadata UID of a Calico v3 resource."""
    try:
        value, _ = _get_with_metadata(resource_kind, namespace, name)
        return value.get("metadata", {}).get("uid", "unknown")
    except etcdv3.KeyNotFound:
        return "unknown"


SANITIZE_LABEL_MAX_LENGTH = 63


def sanitize_label_name_value(name, max_length):
    """Sanitize a label name or value.

    By converting unsupported characters to '_' and ensuring that the first and
    last characters are alphanumeric, and that the length is within a specified
    maximum length.
    """
    name = re.sub("[^-_.A-Za-z0-9]", "_", name[:max_length])
    # Ensure that the first character is alphanumeric, by deleting leading
    # characters that are not alphanumeric.
    m = re.match("^([^A-Za-z0-9]+)", name)
    if m:
        name = name[m.end(1) :]
    # Ensure that the last character is alphanumeric, by deleting trailing
    # characters that are not alphanumeric.
    m = re.match(".*?([^A-Za-z0-9]+)$", name)
    if m:
        name = name[: m.start(1)]

    return name


def _is_namespaced(resource_kind):
    if resource_kind == "WorkloadEndpoint":
        return True
    if resource_kind == "NetworkPolicy":
        return True
    if resource_kind == "LiveMigration":
        return True
    return False


def _plural(resource_kind):
    if resource_kind == "NetworkPolicy":
        return "NetworkPolicies"
    if resource_kind == "GlobalNetworkPolicy":
        return "GlobalNetworkPolicies"
    return resource_kind + "s"


def get_namespace(region_string):
    if region_string is not None and region_string != datamodel_v2.NO_REGION:
        return REGION_NAMESPACE_PREFIX + region_string
    else:
        return NO_REGION_NAMESPACE


def _build_key(resource_kind, namespace, name):
    kind_plural = _plural(resource_kind).lower()
    if _is_namespaced(resource_kind):
        assert namespace is not None
        return "/calico/resources/v3/projectcalico.org/%s/%s/%s" % (
            kind_plural,
            namespace,
            name,
        )
    else:
        return "/calico/resources/v3/projectcalico.org/%s/%s" % (kind_plural, name)


def _get_with_metadata(resource_kind, namespace, name):
    # Note: 'with_metadata' here means including the Calico data model
    # metadata, as well as the etcdv3 mod_revision.
    key = _build_key(resource_kind, namespace, name)
    value_as_string, mod_revision = etcdv3.get(key)
    value = json.loads(value_as_string)
    return value, mod_revision



# Copyright (c) 2015 Cisco Systems.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Import Ml2Plugin before l3_db to fix
# https://github.com/projectcalico/calico/issues/8494
from neutron.plugins.ml2.plugin import Ml2Plugin
from neutron.db import l3_db  # noqa: I100
from neutron.db.models import l3

from neutron_lib import constants

from oslo_config import cfg

from oslo_log import log

from networking_calico.plugins.calico.context import SGRUpdateContext


LOG = log.getLogger(__name__)


class CalicoPlugin(Ml2Plugin, l3_db.L3_NAT_db_mixin):

    # These attributes specify whether the plugin supports or not
    # bulk/pagination/sorting operations. Name mangling is used in
    # order to ensure it is qualified by class
    __native_bulk_support = True
    __native_pagination_support = True
    __native_sorting_support = True

    def __init__(self):
        # Add the ability to handle floating IPs.
        self._supported_extension_aliases.extend(["router"])

        # Suppress the Neutron server's DHCP agent scheduling.  This is useful
        # because it suppresses many WARNING logs that would otherwise appear,
        # but that are actually spurious in a Calico/OpenStack deployment.
        self._supported_extension_aliases.remove("dhcp_agent_scheduler")

        # Set ML2 options so the user doesn't have to.
        LOG.info("Forcing ML2 mechanism_drivers to 'calico'")
        cfg.CONF.set_override("mechanism_drivers", ["calico"], group="ml2")
        LOG.info("Forcing ML2 type_drivers to 'local, flat'")
        cfg.CONF.set_override("type_drivers", ["local", "flat"], group="ml2")
        LOG.info("Forcing ML2 tenant_network_types to 'local'")
        cfg.CONF.set_override("tenant_network_types", ["local"], group="ml2")

        # Here we add, rather than forcing the entire value, because DevStack
        # testing configures 'port-security' here.
        LOG.info("Add 'qos' to ML2 extension_drivers, if not already present")
        if "qos" not in cfg.CONF.ml2.extension_drivers:
            cfg.CONF.set_override(
                "extension_drivers",
                cfg.CONF.ml2.extension_drivers + ["qos"],
                group="ml2",
            )

        # This is a bit of a hack to get the models_v2.Port attributes setup in such
        # a way as to avoid tracebacks in the neutron-server log.
        #
        # The tracebacks are not purely cosmetic as they cause all tests run after
        # TestFloatingIPs to fail, presumably as a result of the neutron-server being
        # left in a bad state. The bad state is likely because the error occurs during a
        # cleanup action ("_clean_garbage") which leaves important resources orphaned
        # with no known way to recover.
        #
        # The side-effects within the Router setup that we care about likely have to do
        # with the "orm.relationship" calls but rather than port those directly here and
        # risk any changes that might occur in the future that could render this
        # assumption false we can just rely on the maintainers of this class to keep
        # things up-to-date on our behalf.
        #
        # An example of the traceback looks like this:
        # Traceback (most recent call last):
        #   .../oslo_service/loopingcall.py", line 150, in _run_loop
        #     result = func(*self.args, **self.kw)
        #   .../neutron/db/l3_db.py", line 163, in _clean_garbage
        #     candidates = self._get_dead_floating_port_candidates(context)
        #   .../neutron/db/l3_db.py", line 198, in _get_dead_floating_port_candidates
        #     return {p['id'] for p in self._core_plugin.get_ports(context, filters)}
        #   .../neutron_lib/db/api.py", line 218, in wrapped
        #     return method(*args, **kwargs)
        #   .../neutron_lib/db/api.py", line 139, in wrapped
        #     setattr(e, '_RETRY_EXCEEDED', True)
        #   .../oslo_utils/excutils.py", line 227, in __exit__
        #     self.force_reraise()
        #   .../oslo_utils/excutils.py", line 200, in force_reraise
        #     raise self.value
        #   .../neutron_lib/db/api.py", line 135, in wrapped
        #     return f(*args, **kwargs)
        #   .../oslo_db/api.py", line 154, in wrapper
        #     ectxt.value = e.inner_exc
        #   .../oslo_utils/excutils.py", line 227, in __exit__
        #     self.force_reraise()
        #   .../oslo_utils/excutils.py", line 200, in force_reraise
        #     raise self.value
        #   .../oslo_db/api.py", line 142, in wrapper
        #     return f(*args, **kwargs)
        #   .../neutron_lib/db/api.py", line 183, in wrapped
        #     LOG.debug("Retry wrapper got retriable exception: %s", e)
        #   .../oslo_utils/excutils.py", line 227, in __exit__
        #     self.force_reraise()
        #   .../oslo_utils/excutils.py", line 200, in force_reraise
        #     raise self.value
        #   .../neutron_lib/db/api.py", line 179, in wrapped
        #     return f(*dup_args, **dup_kwargs)
        #   .../neutron/db/db_base_plugin_v2.py", line 1601, in get_ports
        #     lazy_fields = [models_v2.Port.port_forwardings,
        # AttributeError: type object 'Port' has no attribute 'port_forwardings'
        _ = l3.Router()

        super(CalicoPlugin, self).__init__()

    # Intercept floating IP associates/disassociates so we can trigger an
    # appropriate endpoint update.
    def _update_floatingip(self, context, id, floatingip):
        LOG.info("CalicoPlugin _update_floatingip: %s", floatingip)
        old_floatingip, new_floatingip = super(CalicoPlugin, self)._update_floatingip(
            context, id, floatingip
        )

        LOG.info("CalicoPlugin new_floatingip=%s", new_floatingip)
        if new_floatingip["port_id"]:
            context.fip_update_port_id = new_floatingip["port_id"]
            self.mechanism_manager._call_on_drivers("update_floatingip", context)

        LOG.info("CalicoPlugin old_floatingip=%s", old_floatingip)
        if old_floatingip["port_id"]:
            context.fip_update_port_id = old_floatingip["port_id"]
            self.mechanism_manager._call_on_drivers("update_floatingip", context)

        return old_floatingip, new_floatingip

    def create_floatingip(
        self, context, floatingip, initial_status=constants.FLOATINGIP_STATUS_ACTIVE
    ):
        LOG.info("CalicoPlugin create_floatingip: %s", floatingip)
        new_floatingip = super(CalicoPlugin, self).create_floatingip(
            context, floatingip, initial_status=initial_status
        )
        LOG.info("CalicoPlugin new_floatingip=%s", new_floatingip)
        if new_floatingip["port_id"]:
            context.fip_update_port_id = new_floatingip["port_id"]
            self.mechanism_manager._call_on_drivers("update_floatingip", context)
        return new_floatingip

    def create_security_group_rule(self, context, security_group_rule):
        rule = super().create_security_group_rule(context, security_group_rule)
        sgids = [rule["security_group_id"]]
        self._notify_sg_rule_updated(context, sgids)
        return rule

    def create_security_group_rule_bulk(self, context, security_group_rules):
        rules = super().create_security_group_rule_bulk_native(
            context, security_group_rules
        )
        sgids = set([r["security_group_id"] for r in rules])
        self._notify_sg_rule_updated(context, list(sgids))
        return rules

    def delete_security_group_rule(self, context, sgrid):
        rule = self.get_security_group_rule(context, sgrid)
        super().delete_security_group_rule(context, sgrid)
        self._notify_sg_rule_updated(context, [rule["security_group_id"]])

    def _notify_sg_rule_updated(self, context, sgids):
        self.mechanism_manager._call_on_drivers(
            "security_groups_rule_updated", SGRUpdateContext(context, sgids)
        )



# Copyright 2015 Metaswitch Networks
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import os
import re

import netaddr

from neutron.agent.linux import dhcp

from neutron_lib import constants

from oslo_log import log as logging


LOG = logging.getLogger(__name__)


class DnsmasqRouted(dhcp.Dnsmasq):
    """Dnsmasq DHCP driver for routed virtual interfaces."""

    def __init__(
        self, conf, network, process_monitor, version=None, plugin=None, *args
    ):
        if args:
            super(DnsmasqRouted, self).__init__(
                conf, network, process_monitor, version, plugin, *args
            )
        else:
            super(DnsmasqRouted, self).__init__(
                conf, network, process_monitor, version, plugin
            )
        self.device_manager = CalicoDeviceManager(self.conf, plugin)

    # Frozen copy of Dnsmasq::_build_cmdline_callback from
    # neutron/agent/linux/dhcp.py in Neutron 13.0.2.
    def neutron_13_0_2_build_cmdline_callback(self, pid_file):
        # We ignore local resolv.conf if dns servers are specified
        # or if local resolution is explicitly disabled.
        _no_resolv = (
            "--no-resolv"
            if self.conf.dnsmasq_dns_servers or not self.conf.dnsmasq_local_resolv
            else ""
        )
        cmd = [
            "dnsmasq",
            "--no-hosts",
            _no_resolv,
            "--except-interface=lo",
            "--pid-file=%s" % pid_file,
            "--dhcp-hostsfile=%s" % self.get_conf_file_name("host"),
            "--addn-hosts=%s" % self.get_conf_file_name("addn_hosts"),
            "--dhcp-optsfile=%s" % self.get_conf_file_name("opts"),
            "--dhcp-leasefile=%s" % self.get_conf_file_name("leases"),
            "--dhcp-match=set:ipxe,175",
        ]
        if self.device_manager.driver.bridged:
            cmd += [
                "--bind-interfaces",
                "--interface=%s" % self.interface_name,
            ]
        else:
            cmd += [
                "--bind-dynamic",
                "--interface=%s" % self.interface_name,
                "--interface=tap*",
                "--bridge-interface=%s,tap*" % self.interface_name,
            ]

        possible_leases = 0
        for i, subnet in enumerate(self._get_all_subnets(self.network)):
            mode = None
            # if a subnet is specified to have dhcp disabled
            if not subnet.enable_dhcp:
                continue
            if subnet.ip_version == 4:
                mode = "static"
            else:
                # Note(scollins) If the IPv6 attributes are not set, set it as
                # static to preserve previous behavior
                addr_mode = getattr(subnet, "ipv6_address_mode", None)
                ra_mode = getattr(subnet, "ipv6_ra_mode", None)
                if (
                    addr_mode in [constants.DHCPV6_STATEFUL, constants.DHCPV6_STATELESS]
                    or not addr_mode
                    and not ra_mode
                ):
                    mode = "static"
                if (
                    addr_mode == constants.IPV6_SLAAC
                    and ra_mode == constants.IPV6_SLAAC
                ):
                    mode = "slaac,ra-only"
            cidr = netaddr.IPNetwork(subnet.cidr)

            if self.conf.dhcp_lease_duration == -1:
                lease = "infinite"
            else:
                lease = "%ss" % self.conf.dhcp_lease_duration

            # mode is optional and is not set - skip it
            if mode:
                if subnet.ip_version == 4:
                    cmd.append(
                        "--dhcp-range=%s%s,%s,%s,%s,%s"
                        % (
                            "set:",
                            "subnet-%s" % subnet.id,
                            cidr.network,
                            mode,
                            cidr.netmask,
                            lease,
                        )
                    )
                else:
                    if cidr.prefixlen < 64:
                        LOG.debug(
                            "Ignoring subnet %(subnet)s, CIDR has "
                            "prefix length < 64: %(cidr)s",
                            {"subnet": subnet.id, "cidr": cidr},
                        )
                        continue
                    cmd.append(
                        "--dhcp-range=%s%s,%s,%s,%d,%s"
                        % (
                            "set:",
                            "subnet-%s" % subnet.id,
                            cidr.network,
                            mode,
                            cidr.prefixlen,
                            lease,
                        )
                    )
                possible_leases += cidr.size

        mtu = getattr(self.network, "mtu", 0)
        # Do not advertise unknown mtu
        if mtu > 0:
            cmd.append("--dhcp-option-force=option:mtu,%d" % mtu)

        # Cap the limit because creating lots of subnets can inflate
        # this possible lease cap.
        cmd.append(
            "--dhcp-lease-max=%d" % min(possible_leases, self.conf.dnsmasq_lease_max)
        )

        try:
            if self.conf.dhcp_renewal_time > 0:
                cmd.append(
                    "--dhcp-option-force=option:T1,%ds" % self.conf.dhcp_renewal_time
                )
        except AttributeError:
            pass

        try:
            if self.conf.dhcp_rebinding_time > 0:
                cmd.append(
                    "--dhcp-option-force=option:T2,%ds" % self.conf.dhcp_rebinding_time
                )
        except AttributeError:
            pass

        cmd.append("--conf-file=%s" % self.conf.dnsmasq_config_file)
        for server in self.conf.dnsmasq_dns_servers:
            cmd.append("--server=%s" % server)

        try:
            if self.conf.dns_domain:
                cmd.append("--domain=%s" % self.conf.dns_domain)
        except AttributeError:
            try:
                if self.dns_domain:
                    cmd.append("--domain=%s" % self.dns_domain)
            except AttributeError:
                pass

        if self.conf.dhcp_broadcast_reply:
            cmd.append("--dhcp-broadcast")

        if self.conf.dnsmasq_base_log_dir:
            log_dir = os.path.join(self.conf.dnsmasq_base_log_dir, self.network.id)
            try:
                if not os.path.exists(log_dir):
                    os.makedirs(log_dir)
            except OSError:
                LOG.error("Error while create dnsmasq log dir: %s", log_dir)
            else:
                log_filename = os.path.join(log_dir, "dhcp_dns_log")
                cmd.append("--log-queries")
                cmd.append("--log-dhcp")
                cmd.append("--log-facility=%s" % log_filename)

        return cmd

    def _build_cmdline_callback(self, pid_file):
        cmd = self.neutron_13_0_2_build_cmdline_callback(pid_file)

        # Replace 'static' by 'static,off-link' in all IPv6
        # --dhcp-range options.
        prog = re.compile("(--dhcp-range=set:[^,]+,[0-9a-f:]+),static,(.*)")
        for option in copy.copy(cmd):
            m = prog.match(option)
            if m:
                cmd.remove(option)
                cmd.append(m.group(1) + ",static,off-link," + m.group(2))

        # Add '--enable-ra'.
        cmd.append("--enable-ra")
        # Don't advertise the MTU, since we don't have the value here
        cmd.append("--ra-param=ns-*,mtu:off,0")

        # Enumerate precisely the TAP interfaces to listen on.
        cmd.remove("--interface=tap*")
        cmd.remove("--bridge-interface=%s,tap*" % self.interface_name)
        bridge_option = "--bridge-interface=%s" % self.interface_name
        bridge_ports_added = False
        for port in self.network.ports:
            if port.device_id.startswith("tap"):
                LOG.debug("Listen on %s", port.device_id)
                cmd.append("--interface=%s" % port.device_id)
                bridge_option = bridge_option + "," + port.device_id
                bridge_ports_added = True
        if bridge_ports_added:
            cmd.append(bridge_option)

        return cmd

    def _destroy_namespace_and_port(self):
        try:
            self.device_manager.destroy(self.network, self.interface_name)
        except RuntimeError:
            LOG.warning("Failed trying to delete interface: %s", self.interface_name)


class CalicoDeviceManager(dhcp.DeviceManager):
    """Device manager for the default namespace that Calico operates in."""

    def _set_default_route(self, network, device_name):
        pass

    def _cleanup_stale_devices(self, network, dhcp_port):
        pass

    def fill_dhcp_udp_checksums(self, *args, **kwargs):
        # NOTE(tstachecki): Very old versions of isc-dhcp-client broke when
        # UDP packets had a checksum field not properly filled in as part of
        # GSO/checksum offload support being introduced for virtio_net:
        # https://lwn.net/Articles/373209/
        #
        # The missing/incorrect checksum in the UDP header was problematic for
        # isc-dhcp-client (and possibly other DHCP clients) because they use
        # raw sockets, and were unaware of how to handle
        # TP_STATUS_CSUMNOTREADY.
        #
        # OpenStack has historically worked around this by adding an iptables
        # mangle rule to always rewrite the UDP header for DHCP packets. But,
        # isc-dhcp-client has also been patched to account for this since 2014,
        # and likewise has qemu (for virtio_net) since 2009:
        # https://git.qemu.org/?p=qemu.git;a=commit;h=1d41b0c
        #
        # The consequence of patching the mangle rule in iptables is that the
        # underlying OpenStack mechanisms tend to add the rules in an order
        # that upsets Felix. When Felix sees this (often during live-migration,
        # when the VM has no networking), it can spend a fair amount of time
        # resyncing the rules... thus leaving a live-migrating instance with
        # no networking for an appreciable amount of time.
        #
        # As this should not be needed for any Linux distribution in the last
        # decade and it can perturb Felix, simply do not add in an iptables
        # rule to fill in DHCP UDP checksums here with the assumption that it
        # it has been fixed/is being done elsewhere.
        pass



# -*- coding: utf-8 -*-
# Copyright (c) 2015-2016 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Shared etcd data-model definitions for version 1 of the data model.

The wider Calico data model is now v3.  The legacy v1 definitions here
are private to networking-calico (i.e. the Neutron driver and the
Calico DHCP agent) and Felix's status-reporting code.  However, when
changing these, we still need to consider upgrading an existing
Calico/OpenStack deployment.
"""
from networking_calico.common import intern_string

# Key used for leader election by Neutron mechanism drivers.
NEUTRON_ELECTION_KEY = "/calico/openstack/v1/neutron_election"

ENDPOINT_STATUS_UP = "up"
ENDPOINT_STATUS_DOWN = "down"
ENDPOINT_STATUS_ERROR = "error"

# Subtree used by the Neutron driver to pass subnet information to the
# DHCP agent.
SUBNET_DIR = "/calico/dhcp/v1/subnet"


class EndpointId(object):
    __slots__ = ["host", "endpoint"]

    def __init__(self, host, endpoint):
        # We intern these strings since they can occur in many IDs.  The
        # host and orchestrator are trivially repeated for all endpoints
        # on a host.  The others get repeated over time.
        self.host = intern_string(host)
        self.endpoint = intern_string(endpoint)

    def __str__(self):
        return self.__class__.__name__ + ("<%s>" % self.endpoint)

    def __repr__(self):
        return self.__class__.__name__ + "(%r,%r)" % (self.host, self.endpoint)

    def __ne__(self, other):
        return not (self == other)


class WloadEndpointId(EndpointId):
    __slots__ = ["orchestrator", "workload"]

    def __init__(self, host, orchestrator, workload, endpoint):
        # We intern these strings since they can occur in many IDs.  The
        # host and orchestrator are trivially repeated for all endpoints
        # on a host.  The others get repeated over time.
        super(WloadEndpointId, self).__init__(host, endpoint)
        self.orchestrator = intern_string(orchestrator)
        self.workload = intern_string(workload)

    def __repr__(self):
        return self.__class__.__name__ + "(%r,%r,%r,%r)" % (
            self.host,
            self.orchestrator,
            self.workload,
            self.endpoint,
        )

    def __eq__(self, other):
        if other is self:
            return True
        if not isinstance(other, WloadEndpointId):
            return False
        return (
            other.endpoint == self.endpoint
            and other.workload == self.workload
            and other.host == self.host
            and other.orchestrator == self.orchestrator
        )

    def __hash__(self):
        return hash(self.endpoint) + hash(self.workload)



# -*- coding: utf-8 -*-
# Copyright (c) 2018 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron.db import models_v2
from neutron.db.models.l3 import FloatingIP
from neutron.db.qos import models as qos_models

from neutron_lib import exceptions as n_exc

from oslo_config import cfg

from oslo_log import log

from networking_calico import datamodel_v3
from networking_calico import etcdv3
from networking_calico.common import config as calico_config
from networking_calico.plugins.ml2.drivers.calico.policy import SG_LABEL_PREFIX
from networking_calico.plugins.ml2.drivers.calico.policy import SG_NAME_LABEL_PREFIX
from networking_calico.plugins.ml2.drivers.calico.policy import SG_NAME_MAX_LENGTH
from networking_calico.plugins.ml2.drivers.calico.syncer import ResourceGone
from networking_calico.plugins.ml2.drivers.calico.syncer import ResourceSyncer


LOG = log.getLogger(__name__)


# A lightweight class to hold all of the additional information that we gather
# when translating from a Neutron port to a Calico WorkloadEndpoint.
class PortExtra(object):
    def __init__(self):
        self.fixed_ips = None
        self.floating_ips = None
        self.interface_name = None
        self.network_name = None
        self.project_data = None
        self.qos = None
        self.security_groups = None
        self.security_group_names = {}


# The Calico WorkloadEndpoint that represents an OpenStack VM gets a pair of
# labels to indicate the project (aka tenant) that the VM belongs to.  The
# label names are as follows, and the label values are the actual project ID
# and name at the time of VM creation.
#
# (OpenStack allows a project's name to be updated subsequently; if that
# happens, it is unspecified whether or not we reflect that by updating labels
# of existing WorkloadEndpoints.  In practice the project name label will
# probably change when the WorkloadEndpoint is next rewritten for some other
# reason.  Deployments that use these labels are recommended not to change
# project names post-creation.)
PROJECT_ID_LABEL_NAME = "projectcalico.org/openstack-project-id"
PROJECT_NAME_LABEL_NAME = "projectcalico.org/openstack-project-name"
PROJECT_NAME_MAX_LENGTH = datamodel_v3.SANITIZE_LABEL_MAX_LENGTH
PROJECT_PARENT_ID_LABEL_NAME = "projectcalico.org/openstack-project-parent-id"
NETWORK_NAME_LABEL_NAME = "projectcalico.org/openstack-network-name"
NETWORK_NAME_MAX_LENGTH = datamodel_v3.SANITIZE_LABEL_MAX_LENGTH

# Note: Calico requires a label value to be an empty string, or to consist of
# alphanumeric characters, '-', '_' or '.', starting and ending with an
# alphanumeric character.  If a project name does not already meet that, we
# substitute problem characters so that it does.


class WorkloadEndpointSyncer(ResourceSyncer):

    def __init__(self, db, txn_from_context, policy_syncer, keystone_client):
        super(WorkloadEndpointSyncer, self).__init__(
            db, txn_from_context, "WorkloadEndpoint"
        )
        self.policy_syncer = policy_syncer
        self.keystone = keystone_client
        self.proj_data_cache = {}
        self.region_string = calico_config.get_region_string()
        self.namespace = datamodel_v3.get_namespace(self.region_string)

        # Prime the project data cache now so that we do not pay a fill
        # penalty the first time we need to annotate a port on a cold start.
        self.cache_port_project_data()

    def resync(self, context):
        super(WorkloadEndpointSyncer, self).resync(context)
        self._resync_live_migrations(context)

    def _resync_live_migrations(self, context):
        """Ensure LiveMigration resources and destination WEPs are consistent.

        For each Neutron port with migrating_to set, ensure a destination WEP
        and LiveMigration resource exist.  For each LiveMigration in etcd that
        doesn't correspond to a migrating port, delete it.
        """
        LOG.info("Starting LiveMigration resync")
        namespace = self.namespace

        # Build the set of LiveMigration names that should exist, based on
        # Neutron ports with migrating_to set.
        expected_lm_names = set()
        with self.txn_from_context(context, "resync-live-migrations"):
            for port in self.db.get_ports(context):
                if not _port_is_endpoint_port(port):
                    continue
                dest_host = port.get("binding:profile", {}).get("migrating_to")
                if dest_host is None:
                    continue

                dest_port = port.copy()
                dest_port["binding:host_id"] = dest_host
                dest_wep_name = endpoint_name(dest_port)
                expected_lm_names.add(dest_wep_name)

                # Ensure LiveMigration and destination WEP exist.
                self.write_live_migration(port, dest_port)
                self.write_endpoint(dest_port, context, reread=False)
                LOG.info(
                    "LiveMigration resync: ensured LM and dest WEP for %s",
                    dest_wep_name,
                )

        # Delete orphaned LiveMigration resources.
        for name, _, mod_revision in datamodel_v3.get_all("LiveMigration", namespace):
            if name not in expected_lm_names:
                LOG.warning("LiveMigration resync: deleting orphaned LM %s", name)
                self.delete_live_migration(name, mod_revision=mod_revision)

        LOG.info("LiveMigration resync done")

    def delete_legacy_etcd_data(self):
        if self.namespace != datamodel_v3.NO_REGION_NAMESPACE:
            datamodel_v3.delete_legacy(self.resource_kind, "")

    # The following methods differ from those for other resources because for
    # endpoints we need to read, compare and write labels and annotations as
    # well as spec.

    def get_all_from_etcd(self):
        return datamodel_v3.get_all(
            self.resource_kind, self.namespace, with_labels_and_annotations=True
        )

    def etcd_write_data_matches_existing(self, write_data, existing):
        rspec, rlabels, rannotations = existing
        wspec, wlabels, wannotations = write_data
        return rspec == wspec and rlabels == wlabels and rannotations == wannotations

    def create_in_etcd(self, name, write_data):
        spec, labels, annotations = write_data
        return datamodel_v3.put(
            self.resource_kind,
            self.namespace,
            name,
            spec,
            labels=labels,
            annotations=annotations,
            mod_revision=0,
        )

    def update_in_etcd(self, name, write_data, mod_revision=etcdv3.MUST_UPDATE):
        spec, labels, annotations = write_data
        return datamodel_v3.put(
            self.resource_kind,
            self.namespace,
            name,
            spec,
            labels=labels,
            annotations=annotations,
            mod_revision=mod_revision,
        )

    def delete_from_etcd(self, name, mod_revision):
        return datamodel_v3.delete(
            self.resource_kind, self.namespace, name, mod_revision=mod_revision
        )

    def get_all_from_neutron(self, context):
        # TODO(lukasa): We could reduce the amount of data we load from Neutron
        # here by filtering in the get_ports call.
        return dict(
            (endpoint_name(port), port)
            for port in self.db.get_ports(context)
            if _port_is_endpoint_port(port)
        )

    def neutron_to_etcd_write_data(self, port, context, reread=False):
        if reread:
            try:
                port = self.db.get_port(context, port["id"])
            except n_exc.PortNotFound:
                raise ResourceGone()
        port_extra = self.get_extra_port_information(context, port)
        return (
            endpoint_spec(port, port_extra),
            endpoint_labels(port, self.namespace, port_extra),
            endpoint_annotations(port),
        )

    def write_endpoint(self, port, context, must_update=False, reread=True):
        if reread:
            # Reread the current port. This protects against concurrent writes
            # breaking our state.
            port = self.db.get_port(context, port["id"])

        # Fill out other information we need on the port.
        port_extra = self.get_extra_port_information(context, port)

        # Write the security policies for this port.
        self.policy_syncer.write_sgs_to_etcd(port_extra.security_groups, context)

        # Implementation note: we could arguably avoid holding the transaction
        # for this length and instead release it here, then use atomic CAS. The
        # problem there is that we potentially have to repeatedly respin and
        # regain the transaction. Let's not do that for now, and performance
        # test to see if it's a problem later.
        mod_revision = etcdv3.MUST_UPDATE if must_update else None
        datamodel_v3.put(
            "WorkloadEndpoint",
            self.namespace,
            endpoint_name(port),
            endpoint_spec(port, port_extra),
            labels=endpoint_labels(port, self.namespace, port_extra),
            annotations=endpoint_annotations(port),
            mod_revision=mod_revision,
        )

    def delete_endpoint(self, port):
        return datamodel_v3.delete(
            "WorkloadEndpoint", self.namespace, endpoint_name(port)
        )

    def write_live_migration(self, source_port, dest_port):
        """Write a LiveMigration resource for a migrating port.

        The LiveMigration name is derived from the destination WEP name.
        Returns the UID assigned by etcd (or the existing UID if already
        present).
        """
        namespace = self.namespace
        wep_id_fields = {
            "orchestratorID": "openstack",
            "workloadID": namespace + "/" + source_port["device_id"],
            "endpointID": source_port["id"],
        }
        return datamodel_v3.put(
            "LiveMigration",
            namespace,
            endpoint_name(dest_port),
            {
                "source": {
                    "workloadEndpoint": dict(
                        hostname=source_port["binding:host_id"],
                        **wep_id_fields,
                    ),
                },
                "target": {
                    "workloadEndpoint": dict(
                        hostname=dest_port["binding:host_id"],
                        **wep_id_fields,
                    ),
                },
            },
        )

    def delete_live_migration(self, name, mod_revision=None):
        """Delete a LiveMigration resource by name."""
        return datamodel_v3.delete(
            "LiveMigration", self.namespace, name, mod_revision=mod_revision
        )

    def add_port_interface_name(self, port, port_extra):
        port_extra.interface_name = "tap" + port["id"][:11]

    def get_security_groups_for_port(self, context, port):
        """Checks which security groups apply for a given port.

        Frustratingly, the port dict provided to us when we call get_port may
        actually be out of date, and I don't know why. This change ensures that
        we get the most recent information.
        """
        filters = {"port_id": [port["id"]]}
        bindings = self.db._get_port_security_group_bindings(context, filters=filters)
        return [binding["security_group_id"] for binding in bindings]

    def get_fixed_ips_for_port(self, context, port):
        """Obtains a complete list of fixed IPs for a port.

        Much like with security groups, for some insane reason we're given an
        out of date port dictionary when we call get_port. This forces an
        explicit query of the IPAllocation table to get the right data out of
        Neutron.
        """
        return [
            {"subnet_id": ip["subnet_id"], "ip_address": ip["ip_address"]}
            for ip in context.session.query(models_v2.IPAllocation).filter_by(
                port_id=port["id"]
            )
        ]

    def get_floating_ips_for_port(self, context, port):
        """Obtains a list of floating IPs for a port."""
        return [
            {"int_ip": ip["fixed_ip_address"], "ext_ip": ip["floating_ip_address"]}
            for ip in context.session.query(FloatingIP).filter_by(
                fixed_port_id=port["id"]
            )
        ]

    def get_network_properties_for_port(self, context, port, port_extra):
        network = (
            context.session.query(models_v2.Network)
            .filter_by(id=port["network_id"])
            .first()
        )

        try:
            port_extra.network_name = datamodel_v3.sanitize_label_name_value(
                network["name"],
                NETWORK_NAME_MAX_LENGTH,
            )
        except Exception:
            LOG.warning(f"Failed to find network name for port {port['id']}")

    def get_extra_port_information(self, context, port):
        """get_extra_port_information

        Gets extra information for a port that is needed before sending it to
        etcd.
        """
        LOG.debug("port = %r", port)
        port_extra = PortExtra()
        port_extra.fixed_ips = self.get_fixed_ips_for_port(context, port)
        port_extra.floating_ips = self.get_floating_ips_for_port(context, port)
        port_extra.security_groups = self.get_security_groups_for_port(context, port)
        self.get_network_properties_for_port(context, port, port_extra)

        self.add_port_gateways(context, port_extra)
        self.add_port_interface_name(port, port_extra)
        self.add_port_project_data(port, context, port_extra)
        self.add_port_sg_names(context, port_extra)
        self.add_port_qos(port, context, port_extra)

        return port_extra

    def add_port_gateways(self, context, port_extra):
        """add_port_gateways

        Determine the gateway IP addresses for a given port's IP addresses, and
        adds them to the port dict.

        This method assumes it's being called from within a database
        transaction and does not take out another one.
        """
        for ip in port_extra.fixed_ips:
            subnet = self.db.get_subnet(context, ip["subnet_id"])
            ip["gateway"] = subnet["gateway_ip"]

    def add_port_sg_names(self, context, port_extra):
        """add_port_sg_names

        Determine and store the name of each security group that a port uses.

        This method assumes it's being called from within a database
        transaction and does not take out another one.
        """
        # Oddly, get_security_groups normally tries to create the default SG
        # for the current tenant, and that can hit a
        # NeutronDbObjectDuplicateEntry exception - presumably if there's a
        # race with multiple servers or threads trying to do this at the same
        # time.  Adding "default_sg=True" here suppresses that creation
        # attempt.
        filters = {"id": port_extra.security_groups}
        for sg in self.db.get_security_groups(
            context, filters=filters, default_sg=True
        ):
            sg_name = datamodel_v3.sanitize_label_name_value(
                sg["name"], SG_NAME_MAX_LENGTH
            )
            port_extra.security_group_names[sg["id"]] = sg_name

    def add_port_qos(self, port, context, port_extra):
        """add_port_qos

        Determine and store QoS parameters for a port.

        This method assumes it's being called from within a database
        transaction and does not take out another one.
        """
        qos = {}

        # Minima, maxima and defaults as specified in the WorkloadEndpoint API,
        # and implemented for the Kubernetes case in
        # libcalico-go/lib/backend/k8s/conversion/workload_endpoint_default.go.
        MINMAX_BANDWIDTH = (1000, 10**15)
        MINMAX_BW_BURST = (1000, 34359738360)
        MINMAX_BW_PEAKRATE = (1010, 10**15 + 10**13)
        MINMAX_BW_MINBURST = (1000, 10**8)

        MINMAX_PACKET_RATE = (1, 10**4)
        MINMAX_PR_BURST = (1, 10**4)

        MINMAX_CONNECTIONS = (1, 4294967295)

        def cap(setting, minmax):
            (min, max) = minmax
            if setting < min:
                setting = min
            elif setting > max:
                setting = max
            return setting

        qos_policy_id = port.get("qos_policy_id") or port.get("qos_network_policy_id")
        LOG.debug("QoS Policy ID = %r", qos_policy_id)
        if qos_policy_id:
            rules = context.session.query(qos_models.QosBandwidthLimitRule).filter_by(
                qos_policy_id=qos_policy_id
            )
            for r in rules:
                LOG.debug("BW rule = %r", r)
                direction = r.get("direction", "egress")
                if r["max_kbps"] != 0:
                    qos[direction + "Bandwidth"] = cap(
                        r["max_kbps"] * 1000, MINMAX_BANDWIDTH
                    )
                if r["max_burst_kbps"] != 0:
                    qos[direction + "Peakrate"] = cap(
                        r["max_burst_kbps"] * 1000, MINMAX_BW_PEAKRATE
                    )

            rules = context.session.query(qos_models.QosPacketRateLimitRule).filter_by(
                qos_policy_id=qos_policy_id
            )
            for r in rules:
                LOG.debug("PR rule = %r", r)
                direction = r.get("direction", "egress")
                if r["max_kpps"] != 0:
                    qos[direction + "PacketRate"] = cap(
                        r["max_kpps"] * 1000, MINMAX_PACKET_RATE
                    )

        if cfg.CONF.calico.max_ingress_connections_per_port != 0:
            qos["ingressMaxConnections"] = cap(
                cfg.CONF.calico.max_ingress_connections_per_port, MINMAX_CONNECTIONS
            )
        if cfg.CONF.calico.max_egress_connections_per_port != 0:
            qos["egressMaxConnections"] = cap(
                cfg.CONF.calico.max_egress_connections_per_port, MINMAX_CONNECTIONS
            )

        if "ingressBandwidth" in qos:
            if cfg.CONF.calico.ingress_burst_bits != 0:
                qos["ingressBurst"] = cap(
                    cfg.CONF.calico.ingress_burst_bits, MINMAX_BW_BURST
                )
            else:
                qos["ingressBurst"] = calico_config.DEFAULT_BW_BURST
            if cfg.CONF.calico.ingress_minburst_bytes != 0 and "ingressPeakrate" in qos:
                qos["ingressMinburst"] = cap(
                    cfg.CONF.calico.ingress_minburst_bytes, MINMAX_BW_MINBURST
                )

        if "egressBandwidth" in qos:
            if cfg.CONF.calico.egress_burst_bits != 0:
                qos["egressBurst"] = cap(
                    cfg.CONF.calico.egress_burst_bits, MINMAX_BW_BURST
                )
            else:
                qos["egressBurst"] = calico_config.DEFAULT_BW_BURST
            if cfg.CONF.calico.egress_minburst_bytes != 0 and "egressPeakrate" in qos:
                qos["egressMinburst"] = cap(
                    cfg.CONF.calico.egress_minburst_bytes, MINMAX_BW_MINBURST
                )

        if "ingressPacketRate" in qos:
            if cfg.CONF.calico.ingress_burst_packets != 0:
                qos["ingressPacketBurst"] = cap(
                    cfg.CONF.calico.ingress_burst_packets, MINMAX_PR_BURST
                )
            else:
                qos["ingressPacketBurst"] = calico_config.DEFAULT_PR_BURST

        if "egressPacketRate" in qos:
            if cfg.CONF.calico.egress_burst_packets != 0:
                qos["egressPacketBurst"] = cap(
                    cfg.CONF.calico.egress_burst_packets, MINMAX_PR_BURST
                )
            else:
                qos["egressPacketBurst"] = calico_config.DEFAULT_PR_BURST

        port_extra.qos = qos

    def add_port_project_data(self, port, context, port_extra):
        """add_port_project_data

        Determine the OpenStack project name and parent ID for a given
        port's project/tenant ID, and add it as port_extra.project_data.
        """
        proj_id = port.get("project_id", port.get("tenant_id"))
        if proj_id is None:
            LOG.warning("Port with no project ID: %r", port)
            return

        # If we've already cached the corresponding project data, we're done.
        proj_data = self.proj_data_cache.get(proj_id)
        if proj_data is not None:
            LOG.debug("Project data %r was cached", proj_data)
            port_extra.project_data = proj_data
            return

        # Not cached, so look up the port's project in the Keystone DB.
        self.cache_port_project_data()
        proj_data = self.proj_data_cache.get(proj_id)
        if proj_data is None:
            LOG.warning("Unable to find project data for port: %r", port)
            return

        port_extra.project_data = proj_data

    def cache_port_project_data(self):
        """cache_port_project_data

        Invoked when should populate the project cache for port annotations.
        """
        # Flush the cache if it has reached its maximum allowed size.
        if len(self.proj_data_cache) >= cfg.CONF.calico.project_name_cache_max:
            self.proj_data_cache = {}
        try:
            for proj in self.keystone.projects.list():
                if proj.id not in self.proj_data_cache:
                    LOG.info("Got project name %r from Keystone", proj.name)
                    proj_name = datamodel_v3.sanitize_label_name_value(
                        proj.name, PROJECT_NAME_MAX_LENGTH
                    )
                    self.proj_data_cache[proj.id] = (proj_name, proj.parent_id)
        except Exception:
            # Probably don't have right credentials for that lookup.
            LOG.exception("Failed to query Keystone DB")


def endpoint_name(port):
    def escape_dashes(s):
        return s.replace("-", "--")

    return "%s-openstack-%s-%s" % (
        escape_dashes(port["binding:host_id"]),
        escape_dashes(port["device_id"]),
        escape_dashes(port["id"]),
    )


def endpoint_labels(port, namespace, port_extra):
    labels = {}
    for sg_id in port_extra.security_groups:
        sg_name = port_extra.security_group_names.get(sg_id, "")
        labels[SG_LABEL_PREFIX + sg_id] = sg_name
        if sg_name:
            labels[SG_NAME_LABEL_PREFIX + sg_name] = sg_id
    labels["projectcalico.org/namespace"] = namespace
    labels["projectcalico.org/orchestrator"] = "openstack"

    proj_id = port.get("project_id", port.get("tenant_id"))
    if proj_id is not None:
        labels[PROJECT_ID_LABEL_NAME] = proj_id
    if port_extra.project_data:
        name, parent_id = port_extra.project_data
        labels[PROJECT_NAME_LABEL_NAME] = name
        labels[PROJECT_PARENT_ID_LABEL_NAME] = parent_id

    network_name = port_extra.network_name
    if network_name is not None:
        labels[NETWORK_NAME_LABEL_NAME] = network_name
    return labels


# Represent a Neutron port as a Calico v3 WorkloadEndpoint spec.
def endpoint_spec(port, port_extra):
    """endpoint_spec

    Generate JSON WorkloadEndpointSpec for the given Neutron port.
    """

    # Construct the simpler spec data.
    data = {
        "orchestrator": "openstack",
        "workload": port["device_id"],
        "node": port["binding:host_id"],
        "endpoint": port["id"],
        "interfaceName": port_extra.interface_name,
        "mac": port["mac_address"],
    }

    # Collect IPv4 and IPv6 addresses.  On the way, also set the corresponding
    # gateway fields.  If there is more than one IPv4 or IPv6 gateway, the last
    # one (in port_extra.fixed_ips) wins.
    ip_nets = []
    for ip in port_extra.fixed_ips:
        if ":" in ip["ip_address"]:
            ip_nets.append(ip["ip_address"] + "/128")
            if ip["gateway"] is not None:
                data["ipv6Gateway"] = ip["gateway"]
        else:
            ip_nets.append(ip["ip_address"] + "/32")
            if ip["gateway"] is not None:
                data["ipv4Gateway"] = ip["gateway"]

    # we need to store allowedIPs twice, because
    # dhcp agent creates dhcp record only for fixed IP
    # but felix have to create route for both (fixed and allowed ips)
    allowed_ips = []
    for aap in port.get("allowed_address_pairs", []):
        ip_addr = str(aap["ip_address"])
        if ":" in ip_addr:
            ip_nets.append(ip_addr + "/128")
            allowed_ips.append(ip_addr + "/128")
        else:
            ip_nets.append(ip_addr + "/32")
            allowed_ips.append(ip_addr + "/32")

    data["ipNetworks"] = ip_nets
    data["allowedIps"] = allowed_ips

    ip_nats = []
    for ip in port_extra.floating_ips:
        ip_nats.append(
            {
                "internalIP": ip["int_ip"],
                "externalIP": ip["ext_ip"],
            }
        )
    if ip_nats:
        data["ipNATs"] = ip_nats

    if port_extra.qos:
        data["qosControls"] = port_extra.qos

    # Return that data.
    return data


def endpoint_annotations(port):
    annotations = {datamodel_v3.ANN_KEY_NETWORK_ID: port["network_id"]}

    # If the port has a DNS assignment, represent that as an FQDN annotation.
    dns_assignment = port.get("dns_assignment")
    if dns_assignment:
        # Note: the Neutron server generates a list of assignment entries, one
        # for each fixed IP, but all with the same FQDN, for slightly
        # historical reasons.  We're fine getting the FQDN from the first
        # entry.
        annotations[datamodel_v3.ANN_KEY_FQDN] = dns_assignment[0]["fqdn"]

    return annotations


def _port_is_endpoint_port(port):
    # Return True if port is a VM port.
    if port["device_owner"].startswith("compute:"):
        return True

    # Also return True if port is for a Kuryr container.
    if port["device_owner"].startswith("kuryr:container"):
        return True

    # Otherwise log and return False.
    LOG.debug("Not a VM port: %s" % port)
    return False



# Copyright (c) 2020-2024 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import subprocess
import json
import sys

from tests.k8st.test_base import TestBaseV6
from tests.k8st.utils.utils import start_external_node_with_bgp, \
        retry_until_success, run, curl, DiagsCollector, calicoctl, kubectl, node_info, NGINX_IMAGE

_log = logging.getLogger(__name__)

attempts = 10

bird_conf = """
# Template for all BGP clients
template bgp bgp_template {
  debug { states };
  description "Connection to BGP peer";
  local as 64512;
  multihop;
  gateway recursive; # This should be the default, but just in case.
  import all;        # Import all routes, since we don't know what the upstream
                     # topology is and therefore have to trust the ToR/RR.
  export all;
  source address ip@local;  # The local address we use for the TCP connection
  add paths on;
  graceful restart;  # See comment in kernel section about graceful restart.
  connect delay time 2;
  connect retry time 5;
  error wait time 5,30;
}

# ------------- Node-to-node mesh -------------
protocol bgp Mesh_with_master_node from bgp_template {
  neighbor %s as 64512;
  passive on; # Mesh is unidirectional, peer will connect to us.
}

protocol bgp Mesh_with_node_1 from bgp_template {
  neighbor %s as 64512;
  passive on; # Mesh is unidirectional, peer will connect to us.
}

protocol bgp Mesh_with_node_2 from bgp_template {
  neighbor %s as 64512;
  passive on; # Mesh is unidirectional, peer will connect to us.
}

protocol bgp Mesh_with_node_3 from bgp_template {
  neighbor %s as 64512;
  passive on; # Mesh is unidirectional, peer will connect to us.
}
"""

# BIRD config for an external node to peer with
# the in-cluster route reflector on kube-node-2.
bird_conf_rr = """
# Template for all BGP clients
template bgp bgp_template {
  debug { states };
  description "Connection to BGP peer";
  local as 64512;
  multihop;
  gateway recursive; # This should be the default, but just in case.
  import all;        # Import all routes, since we don't know what the upstream
                     # topology is and therefore have to trust the ToR/RR.
  export all;
  source address ip@local;  # The local address we use for the TCP connection
  add paths on;
  graceful restart;  # See comment in kernel section about graceful restart.
  connect delay time 2;
  connect retry time 5;
  error wait time 5,30;
}

protocol bgp Mesh_with_node_2 from bgp_template {
  neighbor %s as 64512;
  passive on; # Mesh is unidirectional, peer will connect to us.
}
"""

class _TestBGPAdvertV6(TestBaseV6):

    def setUp(self):
        super(_TestBGPAdvertV6, self).setUp()

        # Create bgp test namespace
        self.ns = "bgp-test"
        self.create_namespace(self.ns)

        self.nodes, self.ipv4s, self.ipv6s = node_info()
        self.external_node_ip = start_external_node_with_bgp(
            "kube-node-extra",
            bird6_peer_config=self.get_bird_conf(),
        )

        # Establish BGPPeer from cluster nodes to node-extra
        calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPPeer
metadata:
  name: node-extra.peer%s
EOF
""" % self.get_extra_peer_spec())

    def tearDown(self):
        super(_TestBGPAdvertV6, self).tearDown()
        self.delete_and_confirm(self.ns, "ns")
        try:
            # Delete the extra node.
            run("docker rm -f kube-node-extra")
        except subprocess.CalledProcessError:
            pass

        # Delete BGPPeers.
        calicoctl("delete bgppeer node-extra.peer", allow_fail=True)
        calicoctl("delete bgppeer peer-with-rr", allow_fail=True)

        # Restore node-to-node mesh.
        calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata: {name: default}
spec:
  nodeToNodeMeshEnabled: true
  asNumber: 64512
EOF
""")

        # Remove node-2's route-reflector config, using a retry to avoid
        # transient conflict errors.
        retry_until_success(lambda: self.clear_rr_config(self.nodes[2]))

    def set_rr_config(self, node):
        json_str = calicoctl("get node %s -o json" % node)
        node_dict = json.loads(json_str)
        node_dict['metadata']['labels']['i-am-a-route-reflector'] = 'true'
        node_dict['spec']['bgp']['routeReflectorClusterID'] = '224.0.0.1'
        calicoctl("""apply -f - << EOF
%s
EOF
""" % json.dumps(node_dict))

    def clear_rr_config(self, node):
        json_str = calicoctl("get node %s -o json" % node)
        node_dict = json.loads(json_str)
        node_dict['metadata']['labels'].pop('i-am-a-route-reflector', '')
        node_dict['spec']['bgp'].pop('routeReflectorClusterID', '')
        calicoctl("""apply -f - << EOF
%s
EOF
""" % json.dumps(node_dict))


    def get_svc_cluster_ip(self, svc, ns):
        return kubectl("get svc %s -n %s -o json | jq -r .spec.clusterIP" %
                       (svc, ns)).strip()

    def assert_ecmp_routes(self, dst, via):
        matchStr = dst + " proto bird metric 1024 pref medium"
        # sort ips and construct match string for ECMP routes.
        for ip in sorted(via):
            matchStr += "\n\tnexthop via %s dev eth0 weight 1 " % ip
        retry_until_success(lambda: self.assertIn(matchStr, self.get_routes()))

    def get_svc_host_ipv6(self, svc, ns):
        ipv4 = kubectl("get po -l app=%s -n %s -o json | jq -r .items[0].status.hostIP" %
                       (svc, ns)).strip()
        for i in range(len(self.ipv4s)):
            if ipv4 == self.ipv4s[i]:
                return self.ipv6s[i]
        assert False

    def add_svc_external_ips(self, svc, ns, ips):
        ipsStr = ','.join('"{0}"'.format(ip) for ip in ips)
        patchStr = "{\"spec\": {\"externalIPs\": [%s]}}" % (ipsStr)
        return kubectl("patch svc %s -n %s --patch '%s'" % (svc, ns, patchStr)).strip()


class TestBGPAdvertV6(_TestBGPAdvertV6):

    # In the tests of this class we have a full BGP mesh between the
    # cluster nodes (kube-control-plane, kube-node-1 and kube-node-2)
    # and the external node (kube-node-extra):
    #
    # - The full mesh between the cluster nodes is configured by
    #   nodeToNodeMeshEnabled: true.
    #
    # - The peerings from each cluster node to the external node are
    #   configured by self.get_extra_peer_spec().
    #
    # - The peerings from the external node to each cluster node are
    #   configured by self.get_bird_conf().

    def get_bird_conf(self):
        return bird_conf % (self.ipv6s[0], self.ipv6s[1],
                            self.ipv6s[2], self.ipv6s[3])

    def get_extra_peer_spec(self):
        return """
spec:
  peerIP: %s
  asNumber: 64512
""" % self.external_node_ip

    def test_cluster_ip_advertisement(self):
        """
        Runs the tests for service cluster IPv6 advertisement
        - Create both a Local and a Cluster type NodePort service with a single replica.
          - assert only local and cluster CIDR routes are advertised.
          - assert /128 routes are used, source IP is preserved.
        - Scale the Local NP service so it is running on multiple nodes, assert ECMP routing, source IP is preserved.
        - Delete both services, assert only cluster CIDR route is advertised.
        """
        with DiagsCollector():

            calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata:
  name: default
spec:
  serviceClusterIPs:
  - cidr: fd00:10:96::/112
EOF
""")

            # Assert that a route to the service IP range is present.
            retry_until_success(lambda: self.assertIn("fd00:10:96::/112", self.get_routes()))

            # Create both a Local and a Cluster type NodePort service with a single replica.
            local_svc = "nginx-local"
            cluster_svc = "nginx-cluster"
            self.deploy("gcr.io/kubernetes-e2e-test-images/test-webserver:1.0", local_svc, self.ns, 80, ipv6=True)
            self.deploy("gcr.io/kubernetes-e2e-test-images/test-webserver:1.0", cluster_svc, self.ns, 80, traffic_policy="Cluster", ipv6=True)
            self.wait_until_exists(local_svc, "svc", self.ns)
            self.wait_until_exists(cluster_svc, "svc", self.ns)

            # Get clusterIPs.
            local_svc_ip = self.get_svc_cluster_ip(local_svc, self.ns)
            cluster_svc_ip = self.get_svc_cluster_ip(cluster_svc, self.ns)

            # Wait for the deployments to roll out.
            self.wait_for_deployment(local_svc, self.ns)
            self.wait_for_deployment(cluster_svc, self.ns)

            # Assert that both nginx service can be curled from the external node.
            retry_until_success(curl, function_args=[local_svc_ip])
            retry_until_success(curl, function_args=[cluster_svc_ip])

            # Assert that local clusterIP is an advertised route and cluster clusterIP is not.
            retry_until_success(lambda: self.assertIn(local_svc_ip, self.get_routes()))
            retry_until_success(lambda: self.assertNotIn(cluster_svc_ip, self.get_routes()))

            # TODO: This assertion is actually incorrect. Kubernetes performs
            # SNAT on all traffic destined to a service ClusterIP that doesn't
            # originate from within the cluster's pod CIDR. This assertion
            # pass for External / LoadBalancer IPs, though.
            #
            # Create a network policy that only accepts traffic from the external node.
            # Applying this policy asserts that traffic is not being SNAT'd by kube-proxy
            # when it reaches the destination node.
            # kubectl("""apply -f - << EOF
# apiVersion: networking.k8s.io/v1
# kind: NetworkPolicy
# metadata:
  # name: allow-tcp-80-ex
  # namespace: bgp-test
# spec:
  # podSelector: {}
  # policyTypes:
  # - Ingress
  # ingress:
  # - from:
    # - ipBlock: { cidr: %s/128 }
    # ports:
    # - protocol: TCP
      # port: 80
# EOF
# """ % self.external_node_ip)

            # Connectivity to nginx-local should always succeed.
            for i in range(attempts):
              retry_until_success(curl, retries=200, wait_time=5, function_args=[local_svc_ip])

            # NOTE: Unlike in the IPv6 case (in test_bgp_advert.py) we cannot successfully test that
            # connectivity to nginx-cluster is load-balanced across all nodes (and hence, with the
            # above policy in place, will sometimes fail and sometimes succeed), because our current
            # observation is that Linux's IPv6 ECMP route choice does _not_ depend on source port,
            # even though it is documented as such when fib_multipath_hash_policy == 1.

            # Scale the local_svc to 4 replicas
            self.scale_deployment(local_svc, self.ns, 4)
            self.wait_for_deployment(local_svc, self.ns)
            self.assert_ecmp_routes(local_svc_ip, [self.ipv6s[1], self.ipv6s[2], self.ipv6s[3]])
            for i in range(attempts):
              retry_until_success(curl, function_args=[local_svc_ip])

            # Delete both services.
            self.delete_and_confirm(local_svc, "svc", self.ns)
            self.delete_and_confirm(cluster_svc, "svc", self.ns)

            # Assert that clusterIP is no longer an advertised route.
            retry_until_success(lambda: self.assertNotIn(local_svc_ip, self.get_routes()))

    def test_external_ip_advertisement(self):
        """
        Runs the tests for service external IPv6 advertisement
        """
        with DiagsCollector():

            # Allow two IP ranges for the external IPs we'll test with
            calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata:
  name: default
spec:
  serviceExternalIPs:
  - cidr: fd5f:1234:175:200::/112
  - cidr: fd5f:1234:200:255::/120
EOF
""")

            # Create both a Local and a Cluster type NodePort service with a single replica.
            local_svc = "nginx-local"
            cluster_svc = "nginx-cluster"
            self.deploy("gcr.io/kubernetes-e2e-test-images/test-webserver:1.0", local_svc, self.ns, 80, ipv6=True)
            self.deploy("gcr.io/kubernetes-e2e-test-images/test-webserver:1.0", cluster_svc, self.ns, 80, traffic_policy="Cluster", ipv6=True)
            self.wait_until_exists(local_svc, "svc", self.ns)
            self.wait_until_exists(cluster_svc, "svc", self.ns)

            # Get clusterIPs.
            local_svc_ip = self.get_svc_cluster_ip(local_svc, self.ns)
            cluster_svc_ip = self.get_svc_cluster_ip(cluster_svc, self.ns)

            # Wait for the deployments to roll out.
            self.wait_for_deployment(local_svc, self.ns)
            self.wait_for_deployment(cluster_svc, self.ns)

            # Assert that clusterIPs are not advertised.
            retry_until_success(lambda: self.assertNotIn(local_svc_ip, self.get_routes()))
            retry_until_success(lambda: self.assertNotIn(cluster_svc_ip, self.get_routes()))

            # Create a network policy that only accepts traffic from the external node.
            kubectl("""apply -f - << EOF
apiVersion: networking.k8s.io/v1
kind: NetworkPolicy
metadata:
  name: allow-tcp-80-ex
  namespace: bgp-test
spec:
  podSelector: {}
  policyTypes:
  - Ingress
  ingress:
  - from:
    - ipBlock: { cidr: %s/32 }
    ports:
    - protocol: TCP
      port: 80
EOF
""" % self.external_node_ip)

            # Get host IPs for the nginx pods.
            local_svc_host_ip = self.get_svc_host_ipv6(local_svc, self.ns)
            cluster_svc_host_ip = self.get_svc_host_ipv6(cluster_svc, self.ns)

            # Select an IP from each external IP CIDR.
            local_svc_external_ip = "fd5f:1234:175:200::1"
            cluster_svc_external_ip = "fd5f:1234:200:255::1"

            # Add external IPs to the two services.
            self.add_svc_external_ips(local_svc, self.ns, [local_svc_external_ip])
            self.add_svc_external_ips(cluster_svc, self.ns, [cluster_svc_external_ip])

            # Verify that external IPs for local service is advertised but not the cluster service.
            local_svc_externalips_route = "%s via %s" % (local_svc_external_ip, local_svc_host_ip)
            cluster_svc_externalips_route = "%s via %s" % (cluster_svc_external_ip, cluster_svc_host_ip)
            retry_until_success(lambda: self.assertIn(local_svc_externalips_route, self.get_routes()))
            retry_until_success(lambda: self.assertNotIn(cluster_svc_externalips_route, self.get_routes()))

            # Scale the local_svc to 4 replicas.
            self.scale_deployment(local_svc, self.ns, 4)
            self.wait_for_deployment(local_svc, self.ns)

            # Verify that we have ECMP routes for the external IP of the local service.
            retry_until_success(lambda: self.assert_ecmp_routes(local_svc_external_ip, [self.ipv6s[1], self.ipv6s[2], self.ipv6s[3]]))

            # Delete both services, assert only cluster CIDR route is advertised.
            self.delete_and_confirm(local_svc, "svc", self.ns)
            self.delete_and_confirm(cluster_svc, "svc", self.ns)

            # Assert that external IP is no longer an advertised route.
            retry_until_success(lambda: self.assertNotIn(local_svc_externalips_route, self.get_routes()))

    def test_many_services(self):
        """
        Creates a lot of IPv6 services quickly
        """
        with DiagsCollector():

            calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata:
  name: default
spec:
  serviceClusterIPs:
  - cidr: fd00:10:96::/112
EOF
""")

            # Assert that a route to the service IP range is present.
            retry_until_success(lambda: self.assertIn("fd00:10:96::/112", self.get_routes()))

            # Create a local service and deployment.
            local_svc = "nginx-local"
            self.deploy("gcr.io/kubernetes-e2e-test-images/test-webserver:1.0", local_svc, self.ns, 80, ipv6=True)
            self.wait_for_deployment(local_svc, self.ns)

            # Get clusterIPs.
            cluster_ips = []
            cluster_ips.append(self.get_svc_cluster_ip(local_svc, self.ns))

            # Create many more services which select this deployment.
            num_svc = 300
            for i in range(num_svc):
                name = "nginx-svc-%s" % i
                self.create_service(name, local_svc, self.ns, 80, ipv6=True)

            # Get all of their IPs.
            for i in range(num_svc):
                name = "nginx-svc-%s" % i
                cluster_ips.append(self.get_svc_cluster_ip(name, self.ns))

            # Assert they are all advertised to the other node. This should happen
            # quickly enough that by the time we have queried all services from
            # the k8s API, they should be programmed on the remote node.
            def check_routes_advertised():
                routes = self.get_routes()
                for cip in cluster_ips:
                    self.assertIn(cip, routes)
            retry_until_success(check_routes_advertised, retries=3, wait_time=5)

            # Scale to 0 replicas, assert all routes are removed.
            self.scale_deployment(local_svc, self.ns, 0)
            self.wait_for_deployment(local_svc, self.ns)
            def check_routes_gone():
                routes = self.get_routes()
                for cip in cluster_ips:
                    self.assertNotIn(cip, routes)
            retry_until_success(check_routes_gone, retries=10, wait_time=5)

    def test_bgp_filter_ip_advertisement(self):
        with DiagsCollector():
            # Add BGPConfiguration with serviceClusterIP
            kubectl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata:
  name: default
spec:
  serviceClusterIPs:
  - cidr: fd00:10:96::/112
EOF
""")
            # Assert that a route to the service IP range is present.
            retry_until_success(lambda: self.assertIn("fd00:10:96::/112", self.get_routes()))

            # Create a Local type NodePort service with a single replica.
            local_svc = "nginx-local"
            self.deploy("gcr.io/kubernetes-e2e-test-images/test-webserver:1.0", local_svc, self.ns, 80, ipv6=True)
            self.wait_until_exists(local_svc, "svc", self.ns)

            # Get clusterIPs.
            local_svc_ip = self.get_svc_cluster_ip(local_svc, self.ns)

            # Wait for the deployment to roll out.
            self.wait_for_deployment(local_svc, self.ns)

            # Assert that nginx service can be curled from the external node.
            retry_until_success(curl, function_args=[local_svc_ip])

            # Assert that local clusterIP is an advertised route.
            retry_until_success(lambda: self.assertIn(local_svc_ip, self.get_routes()))

            # Create an export BGP filter that rejects the service IP range
            kubectl("""apply -f - <<EOF
apiVersion: projectcalico.org/v3
kind: BGPFilter
metadata:
  name: test-filter-export-1
spec:
  exportV6:
  - cidr: fd00:10:96::/112
    matchOperator: In
    action: Reject
EOF
""")
        kubectl("patch --type=merge bgppeer node-extra.peer --patch '{\"spec\": {\"filters\": [\"test-filter-export-1\"]}}'")
        self.add_cleanup(lambda: kubectl("patch --type=merge bgppeer node-extra.peer --patch '{\"spec\": {\"filters\": []}}'"))
        self.add_cleanup(lambda: kubectl("delete bgpfilter test-filter-export-1"))

        # Assert that local clusterIP is no longer advertised.
        retry_until_success(lambda: self.assertNotIn(local_svc_ip, self.get_routes()))
        # Assert that a route to the service IP range is no longer present.
        retry_until_success(lambda: self.assertNotIn("fd00:10:96::/112", self.get_routes()))

class TestBGPAdvertV6RR(_TestBGPAdvertV6):

    # In the tests of this class, kube-node-2 acts as an RR, and all
    # the other nodes peer with it.  Here are the peerings that we
    # need for that:
    #
    #                                      RR
    # kube-master     kube-node-1     kube-node-2    kube-node-extra
    #  2001:20::8      2001:20::1      2001:20::2      2001:20::20
    #        |                |         | |    |         |
    #        |                +---------+ |    +---------+
    #        +----------------------------+   Peering -> is configured
    #           These peerings are            by get_extra_peer_spec().
    #           configured by BGPPeer         Peering <- is configured
    #           peer-with-rr                  in get_bird_conf().

    def get_bird_conf(self):
        return bird_conf_rr % self.ipv6s[2]

    def get_extra_peer_spec(self):
        return """
spec:
  node: %s
  peerIP: %s
  asNumber: 64512
""" % (self.nodes[2], self.external_node_ip)

    def test_rr(self):
        # Create ExternalTrafficPolicy Local service with one endpoint on node-1
        kubectl("""apply -f - << EOF
apiVersion: apps/v1
kind: Deployment
metadata:
  name: nginx-rr
  namespace: bgp-test
  labels:
    app: nginx
spec:
  replicas: 1
  selector:
    matchLabels:
      app: nginx
      run: nginx-rr
  template:
    metadata:
      labels:
        app: nginx
        run: nginx-rr
    spec:
      containers:
      - name: nginx-rr
        image: %s
        ports:
        - containerPort: 80
      nodeSelector:
        kubernetes.io/os: linux
        kubernetes.io/hostname: %s
---
apiVersion: v1
kind: Service
metadata:
  name: nginx-rr
  namespace: bgp-test
  labels:
    app: nginx
    run: nginx-rr
spec:
  ipFamilies:
  - IPv6
  externalIPs:
  - fd5f:1234:175:200::1
  ports:
  - port: 80
    targetPort: 80
  selector:
    app: nginx
    run: nginx-rr
  type: NodePort
  externalTrafficPolicy: Local
EOF
""" % (NGINX_IMAGE, self.nodes[1]))

        calicoctl("get nodes -o yaml")
        calicoctl("get bgppeers -o yaml")
        calicoctl("get bgpconfigs -o yaml")

        # Update the node-2 to behave as a route-reflector, using a retry
        # to avoid transient conflict errors.
        retry_until_success(lambda: self.set_rr_config(self.nodes[2]))

        # Disable node-to-node mesh, add cluster and external IP CIDRs to
        # advertise, and configure BGP peering between the cluster nodes and the
        # RR.  (The BGP peering from the external node to the RR is included in
        # get_bird_conf() above.)
        calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata:
  name: default
spec:
  nodeToNodeMeshEnabled: false
  asNumber: 64512
  serviceClusterIPs:
  - cidr: fd00:10:96::/112
  serviceExternalIPs:
  - cidr: fd5f:1234:175:200::/112
EOF
""")

        calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPPeer
metadata: {name: peer-with-rr}
spec:
  peerIP: %s
  asNumber: 64512
EOF
""" % self.ipv6s[2])
        svc_json = kubectl("get svc nginx-rr -n bgp-test -o json")
        svc_dict = json.loads(svc_json)
        cluster_ip = svc_dict['spec']['clusterIP']
        external_ip = svc_dict['spec']['externalIPs'][0]
        retry_until_success(lambda: self.assertIn(cluster_ip, self.get_routes()))
        retry_until_success(lambda: self.assertIn(external_ip, self.get_routes()))


    def test_single_ip_lb_rr(self):
        """
        Tests a /128 LB service with externalTrafficPolicy=Local using a RR
        """
        # Create ExternalTrafficPolicy Local service with one endpoint on node-1
        kubectl("""apply -f - << EOF
apiVersion: apps/v1
kind: Deployment
metadata:
  name: nginx-rr
  namespace: bgp-test
  labels:
    app: nginx
spec:
  replicas: 1
  selector:
    matchLabels:
      app: nginx
      run: nginx-rr
  template:
    metadata:
      labels:
        app: nginx
        run: nginx-rr
    spec:
      containers:
      - name: nginx-rr
        image: %s
        ports:
        - containerPort: 80
      nodeSelector:
        kubernetes.io/os: linux
        kubernetes.io/hostname: %s
---
apiVersion: v1
kind: Service
metadata:
  name: nginx-rr
  namespace: bgp-test
  labels:
    app: nginx
    run: nginx-rr
  annotations:
    metallb.universe.tf/loadBalancerIPs: fdff::96
spec:
  ipFamilies:
  - IPv6
  ports:
  - port: 80
    targetPort: 80
  selector:
    app: nginx
    run: nginx-rr
  type: LoadBalancer
  loadBalancerIP: fdff::96
  externalTrafficPolicy: Local
EOF
""" % (NGINX_IMAGE, self.nodes[1]))

        calicoctl("get nodes -o yaml")
        calicoctl("get bgppeers -o yaml")
        calicoctl("get bgpconfigs -o yaml")

        # Update the node-2 to behave as a route-reflector, using a retry
        # to avoid transient conflict errors.
        retry_until_success(lambda: self.set_rr_config(self.nodes[2]))

        # Disable node-to-node mesh, add cluster and external IP CIDRs to
        # advertise, and configure BGP peering between the cluster nodes and the
        # RR.  (The BGP peering from the external node to the RR is included in
        # get_bird_conf() above.)
        calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata:
  name: default
spec:
  nodeToNodeMeshEnabled: false
  asNumber: 64512
  serviceClusterIPs:
  - cidr: fd00:10:96::/112
  serviceLoadBalancerIPs:
  - cidr: fdff::96/128
EOF
""")

        calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPPeer
metadata: {name: peer-with-rr}
spec:
  peerIP: %s
  asNumber: 64512
EOF
""" % self.ipv6s[2])
        svc_json = kubectl("get svc nginx-rr -n bgp-test -o json")
        svc_dict = json.loads(svc_json)
        cluster_ip = svc_dict['spec']['clusterIP']
        load_balancer_ip = svc_dict['spec']['loadBalancerIP']
        retry_until_success(lambda: self.assertIn(cluster_ip, self.get_routes()))
        retry_until_success(lambda: self.assertIn(load_balancer_ip, self.get_routes()))



#!/usr/bin/env python3
# Copyright (c) 2026 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Strip skipped test cases from JUnit XML files.

Ginkgo marks tests matched by -skip or -skip-package as "skipped" in JUnit
reports. This clutters the Semaphore test results UI with hundreds of entries
that weren't actually run. This script removes those entries in-place.

Usage: strip-skipped-tests.py report/*.xml
"""

import sys
import xml.etree.ElementTree as ET


def strip_skipped(path):
    tree = ET.parse(path)
    root = tree.getroot()
    for suite in root.iter("testsuite"):
        to_remove = [tc for tc in suite.findall("testcase") if tc.find("skipped") is not None]
        if not to_remove:
            continue
        for tc in to_remove:
            suite.remove(tc)
        remaining = suite.findall("testcase")
        suite.set("tests", str(len(remaining)))
        suite.set("skipped", "0")
    tree.write(path, xml_declaration=True, encoding="unicode")


for path in sys.argv[1:]:
    try:
        strip_skipped(path)
    except Exception as e:
        print(f"Warning: failed to process {path}: {e}", file=sys.stderr)
